import ast
from typing import Union, cast

from .object_stream import ObjectStream
from .util_ast import as_ast, function_call


def Range(
    lower_bound: Union[str, int, ast.AST], upper_bound: Union[str, int, ast.AST]
) -> ObjectStream:
    r"""
    Given the lower and upper bound return an object with the range of numbers, similar to python
    range

    Arguments:

        lower_bound     The start of the range
        upper_bound     The end of the range

    Return:

        A new ObjectStream that contains the range of numbers

    """

    return ObjectStream(
        function_call(
            "Range", [cast(ast.AST, as_ast(lower_bound)), cast(ast.AST, as_ast(upper_bound))]
        )
    )



import ast
import sys
from typing import Callable, cast

import pytest

from func_adl.util_ast import (
    _realign_indent,
    as_ast,
    function_call,
    lambda_args,
    lambda_body_replace,
    lambda_build,
    lambda_call,
    lambda_is_identity,
    lambda_is_true,
    lambda_test,
    lambda_unwrap,
    parse_as_ast,
    rewrite_func_as_lambda,
    scan_for_metadata,
)


# Ast parsing
def test_as_ast_integer():
    if sys.version_info < (3, 8):
        assert "Num(n=1)" == ast.dump(as_ast(1))
    elif sys.version_info < (3, 9):
        assert "Constant(value=1, kind=None)" == ast.dump(as_ast(1))
    else:
        assert "Constant(value=1)" == ast.dump(as_ast(1))


def test_as_ast_string():
    if sys.version_info < (3, 8):
        assert "Str(s='hi there')" == ast.dump(as_ast("hi there"))
    elif sys.version_info < (3, 9):
        assert "Constant(value='hi there', kind=None)" == ast.dump(as_ast("hi there"))
    else:
        assert "Constant(value='hi there')" == ast.dump(as_ast("hi there"))


def test_as_ast_string_var():
    s = "hi there"
    if sys.version_info < (3, 8):
        assert "Str(s='hi there')" == ast.dump(as_ast(s))
    elif sys.version_info < (3, 9):
        assert "Constant(value='hi there', kind=None)" == ast.dump(as_ast(s))
    else:
        assert "Constant(value='hi there')" == ast.dump(as_ast(s))


def test_as_ast_list():
    if sys.version_info < (3, 8):
        assert "List(elts=[Str(s='one'), Str(s='two')], ctx=Load())" == ast.dump(
            as_ast(["one", "two"])
        )
    elif sys.version_info < (3, 9):
        assert (
            "List(elts=[Constant(value='one', kind=None), Constant(value='two', "
            "kind=None)], ctx=Load())" == ast.dump(as_ast(["one", "two"]))
        )
    else:
        assert "List(elts=[Constant(value='one'), Constant(value='two')], ctx=Load())" == ast.dump(
            as_ast(["one", "two"])
        )


# Fucntion Calling
def test_function_call_simple():
    a = function_call("dude", [as_ast(1)])
    print(ast.dump(ast.parse("dude(1)")))
    if sys.version_info < (3, 8):
        expected = "Call(func=Name(id='dude', ctx=Load()), args=[Num(n=1)], keywords=[])"
    elif sys.version_info < (3, 9):
        expected = (
            "Call(func=Name(id='dude', ctx=Load()), "
            "args=[Constant(value=1, kind=None)], keywords=[])"
        )
    else:
        expected = "Call(func=Name(id='dude', ctx=Load()), args=[Constant(value=1)], keywords=[])"
    assert expected == ast.dump(a)


# Identity
def test_identity_is():
    assert lambda_is_identity(ast.parse("lambda x: x")) is True


def test_identity_isnot_body():
    assert lambda_is_identity(ast.parse("lambda x: x+1")) is False


def test_identity_isnot_args():
    assert lambda_is_identity(ast.parse("lambda x,y: x")) is False


def test_identity_isnot_body_var():
    assert lambda_is_identity(ast.parse("lambda x: x1")) is False


# Is this a lambda?
def test_lambda_test_expression():
    assert lambda_test(ast.parse("x")) is False


def test_lambda_assure_expression():
    try:
        lambda_test(ast.parse("x"))
        assert False
    except Exception:
        pass


def test_lambda_assure_lambda():
    try:
        lambda_test(ast.parse("lambda x : x+1"))
        assert False
    except Exception:
        pass


def test_lambda_args():
    args = lambda_args(ast.parse("lambda x: x+1"))
    assert len(args.args) == 1
    assert args.args[0].arg == "x"


def test_lambda_simple_ast_expr():
    assert lambda_test(ast.Not()) is False


def test_lambda_build_single_arg():
    expr = ast.parse("x+1")
    ln = lambda_build("x", expr)
    assert isinstance(ln, ast.Lambda)


def test_lambda_build_list_arg():
    expr = ast.parse("x+1")
    ln = lambda_build(["x"], expr)
    assert isinstance(ln, ast.Lambda)


def test_lambda_build_proper():
    "Make sure we are building the ast right for the version of python we are in"
    expr = ast.parse("x+1").body[0].value  # type: ignore
    ln = lambda_build("x", expr)
    assert ast.dump(ast.parse("lambda x: x+1").body[0].value) == ast.dump(ln)  # type: ignore


def test_call_wrap_list_arg():
    ln = ast.parse("lambda x: x+1")
    c = lambda_call(["x"], ln)
    assert isinstance(c, ast.Call)


def test_call_wrap_single_arg():
    ln = ast.parse("lambda x: x+1")
    c = lambda_call("x", ln)
    assert isinstance(c, ast.Call)


def test_lambda_test_lambda_module():
    assert lambda_test(ast.parse("lambda x: x")) is True


def test_lambda_test_raw_lambda():
    rl = cast(ast.Expr, ast.parse("lambda x: x").body[0]).value
    assert lambda_test(rl) is True


# Is this lambda always returning true?
def test_lambda_is_true_yes():
    assert lambda_is_true(ast.parse("lambda x: True")) is True


def test_lambda_is_true_no():
    assert lambda_is_true(ast.parse("lambda x: False")) is False


def test_lambda_is_true_expression():
    assert lambda_is_true(ast.parse("lambda x: x")) is False


def test_lambda_is_true_non_lambda():
    assert lambda_is_true(ast.parse("True")) is False


# Replacement
def test_lambda_replace_simple_expression():
    a1 = ast.parse("lambda x: x")

    nexpr = ast.parse("lambda y: y + 1")
    expr = lambda_unwrap(nexpr).body

    a2 = lambda_body_replace(lambda_unwrap(a1), expr)
    a2_txt = ast.dump(a2)
    if sys.version_info < (3, 8):
        assert "op=Add(), right=Num(n=1))" in a2_txt
    elif sys.version_info < (3, 9):
        assert "op=Add(), right=Constant(value=1, kind=None))" in a2_txt
    else:
        assert "op=Add(), right=Constant(value=1))" in a2_txt


def test_rewrite_oneliner():
    a = ast.parse(
        """def oneline(a):
        return a+1"""
    )

    b = a.body[0]
    assert isinstance(b, ast.FunctionDef)
    ln = rewrite_func_as_lambda(b)

    assert isinstance(ln, ast.Lambda)
    assert len(ln.args.args) == 1
    assert ln.args.args[0].arg == "a"
    assert isinstance(ln.body, ast.BinOp)


def test_rewrite_twoliner():
    a = ast.parse(
        """def oneline(a):
        t = a+1
        return t"""
    )

    b = a.body[0]
    assert isinstance(b, ast.FunctionDef)
    with pytest.raises(ValueError) as e:
        rewrite_func_as_lambda(b)

    assert "simple" in str(e.value)


def test_rewrite_noret():
    a = ast.parse(
        """def oneline(a):
        a+1"""
    )

    b = a.body[0]
    assert isinstance(b, ast.FunctionDef)
    with pytest.raises(ValueError) as e:
        rewrite_func_as_lambda(b)

    assert "return" in str(e.value)


def test_parse_as_ast_lambda():
    ln = lambda_unwrap(ast.parse("lambda x: x + 1"))
    r = parse_as_ast(ln)
    assert isinstance(r, ast.Lambda)


def test_parse_as_str():
    r = parse_as_ast("lambda x: x + 1")
    assert isinstance(r, ast.Lambda)


def test_parse_as_callable_simple():
    r = parse_as_ast(lambda x: x + 1)
    assert isinstance(r, ast.Lambda)


def test_parse_nested_lambda():
    r = parse_as_ast(lambda x: (lambda y: y + 1)(x))
    assert isinstance(r, ast.Lambda)
    assert isinstance(r.body, ast.Call)


def test_parse_lambda_capture():
    cut_value = 30
    r = parse_as_ast(lambda x: x > cut_value)
    r_true = parse_as_ast(lambda x: x > 30)
    assert ast.dump(r) == ast.dump(r_true)


def test_parse_lambda_capture_ignore_local():
    x = 30  # NOQA type: ignore
    r = parse_as_ast(lambda x: x > 20)
    r_true = parse_as_ast(lambda y: y > 20)
    assert ast.dump(r) == ast.dump(r_true).replace("'y'", "'x'")


g_cut_value = 30


def test_parse_lambda_capture_ignore_global():
    x = 30  # NOQA type: ignore
    r = parse_as_ast(lambda g_cut_value: g_cut_value > 20)
    r_true = parse_as_ast(lambda y: y > 20)
    assert ast.dump(r) == ast.dump(r_true).replace("'y'", "'g_cut_value'")


def test_parse_lambda_capture_nested_global():
    r = parse_as_ast(lambda x: (lambda y: y > g_cut_value)(x))
    r_true = parse_as_ast(lambda x: (lambda y: y > 30)(x))
    assert ast.dump(r) == ast.dump(r_true)


def test_parse_lambda_capture_nested_local():
    cut_value = 30
    r = parse_as_ast(lambda x: (lambda y: y > cut_value)(x))
    r_true = parse_as_ast(lambda x: (lambda y: y > 30)(x))
    assert ast.dump(r) == ast.dump(r_true)


def test_parse_simple_func():
    "A oneline function defined at local scope"

    def doit(x):
        return x + 1

    f = parse_as_ast(doit)

    assert isinstance(f, ast.Lambda)
    assert len(f.args.args) == 1
    assert isinstance(f.body, ast.BinOp)


def global_doit(x):
    return x + 1


def test_parse_global_simple_func():
    "A oneline function defined at global scope"

    f = parse_as_ast(global_doit)

    assert isinstance(f, ast.Lambda)
    assert len(f.args.args) == 1
    assert isinstance(f.body, ast.BinOp)


g_val = 50


def global_doit_capture(x):
    return x + g_val


def global_doit_capture_true(x):
    return x + 50


def test_parse_global_capture():
    "Global function, which includes variable capture"
    f = parse_as_ast(global_doit_capture)
    f_true = parse_as_ast(global_doit_capture_true)
    assert ast.dump(f) == ast.dump(f_true)


def test_unknown_function():
    "function that isn't declared"
    f = parse_as_ast(lambda a: unknown(a))  # type: ignore # NOQA
    assert "Name(id='unknown'" in ast.dump(f)


def test_known_local_function():
    "function that is declared locally"

    def doit(x):
        ...

    f = parse_as_ast(lambda a: doit(a))  # type: ignore # NOQA
    assert "Name(id='doit'" in ast.dump(f)


def global_doit_non_func(x):
    ...


def test_known_global_function():
    "function that is declared locally"

    f = parse_as_ast(lambda a: global_doit_non_func(a))  # type: ignore # NOQA
    assert "Name(id='global_doit_non_func'" in ast.dump(f)


def test_lambda_args_differentiation():
    "Use the arguments of the lambda to tell what we want"
    found = []

    class my_obj:
        def do_it(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

    (my_obj().do_it(lambda x: x + 1).do_it(lambda y: y * 2))

    assert len(found) == 2
    l1, l2 = found
    assert isinstance(l1, ast.Lambda)
    assert isinstance(l1.body, ast.BinOp)
    assert isinstance(l1.body.op, ast.Add)

    assert isinstance(l2, ast.Lambda)
    assert isinstance(l2.body, ast.BinOp)
    assert isinstance(l2.body.op, ast.Mult)


def test_lambda_method_differentiation():
    "Use the method to tell what we want"
    found1 = []
    found2 = []

    class my_obj:
        def do_it1(self, x: Callable):
            found1.append(parse_as_ast(x, caller_name="do_it1"))
            return self

        def do_it2(self, x: Callable):
            found2.append(parse_as_ast(x, caller_name="do_it2"))
            return self

    (my_obj().do_it1(lambda x: x + 1).do_it2(lambda x: x * 2))

    assert len(found1) == 1
    assert len(found2) == 1

    l1 = found1[0]
    l2 = found2[0]

    assert isinstance(l1, ast.Lambda)
    assert isinstance(l1.body, ast.BinOp)
    assert isinstance(l1.body.op, ast.Add)

    assert isinstance(l2, ast.Lambda)
    assert isinstance(l2.body, ast.BinOp)
    assert isinstance(l2.body.op, ast.Mult)


# def test_parse_continues_accross_lines():
# NOTE: This test will fail because the python tokenizer will treat
# the continuation as a single line, and the parser will see these on
# the same line. This is how python works, and we should follow it.
#     "Use a line continuation and make sure we can tell the difference"
#     found = []

#     class my_obj:
#         def do_it(self, x: Callable):
#             found.append(parse_as_ast(x))
#             return self

#     # fmt: off
#     my_obj().do_it(lambda x: x + 1) \
#         .do_it(lambda x: x * 2)
#     # fmt: on

#     assert len(found) == 2
#     l1, l2 = found
#     assert isinstance(l1, ast.Lambda)
#     assert isinstance(l1.body, ast.BinOp)
#     assert isinstance(l1.body.op, ast.Add)

#     assert isinstance(l2, ast.Lambda)
#     assert isinstance(l2.body, ast.BinOp)
#     assert isinstance(l2.body.op, ast.Mult)


def test_decorator_parse():
    "More general case"

    seen_lambdas = []

    def dec_func(x: Callable):
        def make_it(y: Callable):
            return y

        seen_lambdas.append(parse_as_ast(x))
        return make_it

    @dec_func(lambda y: y + 2)
    def doit(x):
        return x + 1

    assert len(seen_lambdas) == 1
    l1 = seen_lambdas[0]
    assert isinstance(l1.body, ast.BinOp)
    assert isinstance(l1.body.op, ast.Add)


def test_indent_parse():
    "More general case"

    seen_funcs = []

    class h:
        @staticmethod
        def dec_func(x: Callable):
            def make_it(y: Callable):
                return y

            seen_funcs.append(x)
            return make_it

    class yo_baby:
        @h.dec_func(lambda y: y + 2)
        def doit(self, x: int):
            ...

    assert len(seen_funcs) == 1
    l1 = parse_as_ast(seen_funcs[0], "dec_func")
    assert isinstance(l1.body, ast.BinOp)
    assert isinstance(l1.body.op, ast.Add)


def test_two_deep_parse():
    "More general case"

    seen_lambdas = []

    def func_bottom(x: Callable):
        seen_lambdas.append(parse_as_ast(x))

    def func_top(x: Callable):
        func_bottom(x)

    func_top(lambda x: x + 1)

    assert len(seen_lambdas) == 1
    l1 = seen_lambdas[0]
    assert isinstance(l1.body, ast.BinOp)
    assert isinstance(l1.body.op, ast.Add)


def test_parse_continues_one_line():
    "Make sure we do not let our confusion confuse the user - bomb correctly here"
    found = []

    class my_obj:
        def do_it(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

    with pytest.raises(Exception) as e:
        my_obj().do_it(lambda x: x + 1).do_it(lambda x: x * 2)

    assert "multiple" in str(e.value)


def test_parse_space_after_method():
    "Make sure we correctly parse a funnily formatted lambda"

    found = []

    class my_obj:
        def do_it(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

    # fmt: off
    my_obj().do_it   (lambda x: x + 123)  # noqa: E211
    # fmt: on

    assert "123" in ast.dump(found[0])


def test_parse_multiline_bad_line_break():
    "Make sure we correctly parse a funnily formatted lambda"

    found = []

    class my_obj:
        def do_it(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

    # fmt: off
    my_obj().do_it(
        lambda x: x +  # noqa: W504
        123)
    # fmt: on

    assert "123" in ast.dump(found[0])


def test_parse_multiline_lambda_ok_with_one():
    "Make sure we can properly parse a multi-line lambda"

    found = []

    class my_obj:
        def do_it(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

    # fmt: off
    my_obj().do_it(
        lambda x: x
        + 1  # noqa: W503
        + 2  # noqa: W503
        + 20  # noqa: W503
    )
    # fmt: on

    assert "20" in ast.dump(found[0])


def test_parse_multiline_lambda_same_line():
    "Make sure we can properly parse a multi-line lambda"

    found = []

    class my_obj:
        def do_it(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

    # fmt: off
    my_obj().do_it(lambda x: x
                   + 1  # noqa: W503
                   + 2  # noqa: W503
                   + 20  # noqa: W503
                   )
    # fmt: on

    assert "20" in ast.dump(found[0])


def test_parse_multiline_lambda_ok_with_one_and_paran():
    "Make sure we can properly parse a multi-line lambda - using parens as delimiters"

    found = []

    class my_obj:
        def do_it(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

    # fmt: off
    my_obj().do_it(
        lambda x: (
            x
            + 1  # noqa: W503
            + 2  # noqa: W503
            + 20  # noqa: W503
        )
    )
    # fmt: on

    assert "20" in ast.dump(found[0])


def test_parse_multiline_lambda_blank_lines_no_infinite_loop():
    "Make sure we can properly parse a multi-line lambda - using parens as delimiters"

    found = []

    class my_obj:
        def Where(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

        def Select(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

        def AsAwkwardArray(self, stuff: str):
            return self

        def value(self):
            return self

    jets_pflow_name = "hi"
    ds_dijet = my_obj()

    # fmt: off
    jets_pflow = (
        ds_dijet.Select(lambda e: e.Jets(uncalibrated_collection=jets_pflow_name))
        .Select(lambda e: e.Where(lambda j: (j.pt() / 1000) > 30))
        .Select(lambda e: e.Select(lambda j: (j.pt() / 1000)))
        .AsAwkwardArray("JetPt")
        .value()
    )
    # fmt: on
    assert jets_pflow is not None  # Just to keep flake8 happy without adding a noqa above.
    assert "uncalibrated_collection" in ast.dump(found[0])


def test_parse_multiline_lambda_ok_with_one_as_arg():
    "Make sure we can properly parse a multi-line lambda - but now with argument"

    found = []

    class my_obj:
        def do_it(self, x: Callable, counter: int):
            found.append(parse_as_ast(x))
            assert counter > 0
            return self

    # fmt: off
    my_obj().do_it(
        lambda x: x
        + 1  # noqa: W503
        + 2  # noqa: W503
        + 20,  # noqa: W503
        50,
    )
    # fmt: on

    assert "20" in ast.dump(found[0])


def test_parse_multiline_lambda_with_funny_split():
    "This isn't on two lines but sort-of is - so we should parse it. See issue #84"

    found = []

    class my_obj:
        def do_it(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

    # fmt: off
    my_obj().do_it(lambda event: event + 1
                   ).do_it(lambda event: event)
    # fmt: on

    assert "Add()" in ast.dump(found[0])
    assert "Add()" not in ast.dump(found[1])


def test_parse_multiline_lambda_with_comment():
    "Comment in the middle of things"

    found = []

    class my_obj:
        def Where(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

        def Select(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

        def AsAwkwardArray(self, stuff: str):
            return self

        def value(self):
            return self

    source = my_obj()
    # fmt: off
    # flake8: noqa
    r = source.Where(lambda e:
        e.electron_pt.Where(lambda pT: pT > 25).Count() + e.muon_pt.Where(lambda pT: pT > 25).Count()== 1) \
        .Where(lambda e:\
            e.jet_pt.Where(lambda pT: pT > 25).Count() >= 4
        )     # noqa: E501
    # fmt: on

    assert "electron_pt" in ast.dump(found[0])


def test_parse_black_split_lambda_funny():
    "Seen in wild - formatting really did a number here"

    found = []

    class my_obj:
        def do_it(self, x: Callable):
            found.append(parse_as_ast(x))
            return self

    # fmt: off
    my_obj().do_it(
        lambda e: e.Jets("AntiKt4EMTopoJets").do_it(
            lambda j: j.Jets("AntiKt4EMTopoJets").do_it(
                lambda j1: j1.pt() / 1000.0
            )
        )
    )
    # fmt: on

    assert len(found) == 1
    assert "AntiKt4EMTopoJets" in ast.dump(found[0])


def test_parse_metadata_there():
    recoreded = None

    def callback(a: ast.arg):
        nonlocal recoreded
        recoreded = a

    scan_for_metadata(ast.parse("MetaData(e, 22)"), callback)

    assert recoreded is not None
    assert 22 == ast.literal_eval(recoreded)


def test_realign_no_indent():
    assert _realign_indent("test") == "test"


def test_realign_indent_sp():
    assert _realign_indent("    test") == "test"


def test_realign_indent_tab():
    assert _realign_indent("\ttest") == "test"


def test_realign_indent_2lines():
    assert _realign_indent("    test()\n    dude()") == "test()\ndude()"


def test_realign_indent_2lines_uneven():
    assert _realign_indent("    test()\n        dude()") == "test()\n    dude()"



from __future__ import annotations

import ast
import copy
import logging
from typing import Any, Callable, Iterable, Tuple, Type, TypeVar, cast

import pytest

from func_adl import ObjectStream
from func_adl.type_based_replacement import (
    func_adl_callable,
    func_adl_callback,
    remap_by_types,
    remap_from_lambda,
)

#
# NOTE: Keep the tests here the same as in the file `test_type_based_replacement`.
# When 3.11 is the lowest version this file can be deleted.
# THis makes sure that delayed type-hint resolution will work properly!
#


def add_track_extra_info(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
    s_update = s.MetaData({"t": "track stuff"})
    return s_update, a


@func_adl_callback(add_track_extra_info)
class TrackStuff:
    def pt(self) -> float:
        ...

    def eta(self) -> float:
        ...


class Track:
    def pt(self) -> float:
        ...

    def eta(self) -> float:
        ...


class Jet:
    def pt(self) -> float:
        ...

    def eta(self) -> float:
        ...

    def tracks(self) -> Iterable[Track]:
        ...


def ast_lambda(lambda_func: str) -> ast.Lambda:
    "Return the ast starting from the Lambda node"
    return ast.parse(lambda_func).body[0].value  # type: ignore


T = TypeVar("T")


def add_met_extra_info(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
    s_update = s.MetaData({"j": "pxyz stuff"})
    return s_update, a


@func_adl_callback(add_met_extra_info)
class met_extra:
    def pxy(self) -> float:
        ...


def add_met_info(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
    s_update = s.MetaData({"j": "pxy stuff"})
    return s_update, a


def add_met_method_info(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
    s_update = s.MetaData({"j": "custom stuff"})
    return s_update, a


@func_adl_callback(add_met_info)
class met:
    def pxy(self) -> float:
        ...

    def isGood(self) -> bool:
        ...

    def metobj(self) -> met_extra:
        ...

    @func_adl_callback(add_met_method_info)
    def custom(self) -> float:
        ...


def add_collection(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
    """Add a collection to the object stream"""
    assert isinstance(a.func, ast.Attribute)
    if a.func.attr == "Jets":
        s_update = s.MetaData({"j": "stuff"})
        return s_update, a
    elif a.func.attr == "EventNumber":
        new_call = copy.copy(a)
        new_call.args = [ast_lambda("20")]
        return s, new_call
    else:
        return s, a


@func_adl_callback(add_collection)
class Event:
    def Jets(self, bank: str = "default") -> Iterable[Jet]:
        ...

    def Jets_req(self, bank_required: str) -> Iterable[Jet]:
        ...

    def MET(self) -> met:
        ...

    def MET_noreturntype(self):
        ...

    def Tracks(self) -> Iterable[Track]:
        ...

    def TrackStuffs(self) -> Iterable[TrackStuff]:
        ...

    def EventNumber(self) -> int:
        ...


def return_type_test(expr: str, arg_type: type, expected_type: type):
    s = ast_lambda(expr)
    objs = ObjectStream(ast.Name(id="e", ctx=ast.Load()), arg_type)

    _, _, expr_type = remap_by_types(objs, "e", arg_type, s)
    assert expr_type == expected_type


def test_int():
    return_type_test("1", int, int)


def test_int_neg():
    return_type_test("-11", int, int)


def test_bool():
    return_type_test("False", int, bool)


def test_str():
    return_type_test('"hi"', int, str)


def test_float():
    return_type_test("1.5", int, float)


def test_any():
    return_type_test("e", Any, Any)


def test_neg_float():
    return_type_test("-1.5", int, float)


def test_add_int():
    return_type_test("e+1", int, int)


def test_add_float():
    return_type_test("e+1.5", int, float)


def test_sub_int():
    return_type_test("e-1", int, int)


def test_sub_float():
    return_type_test("e-1.5", int, float)


def test_mul_int():
    return_type_test("e*1", int, int)


def test_mul_float():
    return_type_test("e*1.5", int, float)


def test_div_int():
    return_type_test("e/2", int, float)


def test_dib_float():
    return_type_test("e/1.5", int, float)


def test_bool_expression():
    "A bool expression"
    return_type_test("1 > 2", int, bool)


def test_abs_function_int_e():
    "A call to abs with an integer"
    return_type_test("abs(e)", int, float)


def test_abs_function_int_const():
    "A call to abs with an integer"
    return_type_test("abs(-23)", int, float)


def test_abs_function_float():
    "A call to abs with an float"
    return_type_test("abs(e)", float, float)


def test_ifexpr_onetype():
    "A ? expression"
    return_type_test("1 if True else 2", int, int)


def test_ifexpr_onetype_twotype_math():
    "A ? expression"
    return_type_test("1 if True else 2.2", int, float)


def test_ifexpr_onetype_twotypes():
    "A ? expression"
    with pytest.raises(ValueError) as e:
        return_type_test('1 if True else "2.2"', int, float)
    assert "str" in str(e)


def test_subscript():
    return_type_test("e[0]", Iterable[int], int)


def test_subscript_any():
    return_type_test("e[0]", Any, Any)


def test_unknown_name(caplog):
    return_type_test("e+cpp_any", Any, Any)

    assert "cpp_any" in caplog.text


def test_collection():
    "A simple collection"
    s = ast_lambda("e.Jets('default')")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.Jets('default')"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'stuff'})"))
    assert expr_type == Iterable[Jet]


def test_required_arg():
    "A simple collection"
    s = ast_lambda("e.Jets_req()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    with pytest.raises(ValueError) as e:
        remap_by_types(objs, "e", Event, s)

    assert "bank_required" in str(e)


def test_collection_with_default():
    "A simple collection"
    s = ast_lambda("e.Jets()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.Jets('default')"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'stuff'})"))
    assert expr_type == Iterable[Jet]


def test_shortcut_2nested_callback():
    """When there is a simple return, like Where, make sure that lambdas
    inside the method are called, but double inside"""

    s = ast_lambda(
        "ds.Select(lambda e: e.TrackStuffs()).Select(lambda ts: ts.Where(lambda t: t.pt() > 10))"
    )
    objs = ObjectStream[Iterable[Event]](ast.Name(id="ds", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "ds", Iterable[Event], s)

    assert ast.dump(new_s) == ast.dump(
        ast_lambda(
            "ds.Select(lambda e: e.TrackStuffs())"
            ".Select(lambda ts: ts.Where(lambda t: t.pt() > 10))"
        )
    )
    assert ast.dump(new_objs.query_ast) == ast.dump(
        ast_lambda("MetaData(ds, {'t': 'track stuff'})")
    )
    assert expr_type == Iterable[Iterable[TrackStuff]]


R = TypeVar("R")
Result = TypeVar("Result")


class _itsb_FADLStream(Iterable[R]):
    def Select(self, x: Callable[[R], Result]) -> _itsb_FADLStream[Result]:
        ...


class _itsb_MyTrack:
    def pt(self) -> float:
        ...


def test_shortcut_nested_with_iterable_subclass():
    """When there is a simple return, like Where, make sure that lambdas
    inside the method are called"""

    class MyEvent:
        def MyTracks(self) -> _itsb_FADLStream[_itsb_MyTrack]:
            ...

    s = ast_lambda(
        "ds.Select(lambda e: e.MyTracks()).Select(lambda ts: ts.Select(lambda t: t.pt()))"
    )
    objs = ObjectStream[Iterable[MyEvent]](ast.Name(id="ds", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "ds", Iterable[MyEvent], s)

    assert ast.dump(new_s) == ast.dump(
        ast_lambda(
            "ds.Select(lambda e: e.MyTracks()).Select(lambda ts: ts.Select(lambda t: t.pt()))"
        )
    )
    # assert ast.dump(new_objs.query_ast) == ast.dump(
    #     ast_lambda("MetaData(e, {'t': 'track stuff'})")
    # )
    assert expr_type == Iterable[Iterable[float]]


def test_method_on_collection():
    "Call a method that requires some special stuff on a returend object"
    s = ast_lambda("e.MET().pxy()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET().pxy()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'pxy stuff'})"))
    assert expr_type == float


def test_method_callback():
    "Call a method that requires some special stuff on a returend object"
    s = ast_lambda("e.MET().custom()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET().custom()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(
        ast_lambda("MetaData(MetaData(e, {'j': 'pxy stuff'}), {'j': 'custom stuff'})")
    )
    assert expr_type == float


def test_method_on_collection_bool():
    "Call a method that requires some special stuff on a returend object"
    s = ast_lambda("e.MET().isGood()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    _, _, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == bool


def test_method_on_method_on_collection():
    "Call a method that requires some special stuff on a returend object"
    s = ast_lambda("e.MET().metobj().pxy()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()), Event)

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET().metobj().pxy()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(
        ast_lambda("MetaData(MetaData(e, {'j': 'pxy stuff'}), {'j': 'pxyz stuff'})")
    )
    assert expr_type == float


def test_method_modify_ast():
    "Call a method that requires some special stuff on a returend object"
    s = ast_lambda("e.EventNumber()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.EventNumber(20)"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == int


def test_method_with_no_return_type(caplog):
    "A simple collection"
    caplog.set_level(logging.WARNING)
    s = ast_lambda("e.MET_noreturntype().pxy()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET_noreturntype().pxy()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == Any
    assert "MET_noreturntype" in caplog.text


def test_method_with_no_prototype(caplog):
    "A simple collection"
    caplog.set_level(logging.WARNING)
    s = ast_lambda("e.MET_bogus().pxy()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET_bogus().pxy()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == Any
    assert "MET_bogus" in caplog.text


def test_method_with_no_inital_type(caplog):
    "A simple collection"
    caplog.set_level(logging.WARNING)
    s = ast_lambda("e.MET_bogus().pxy()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Any, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET_bogus().pxy()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == Any
    assert len(caplog.text) == 0


def test_bogus_method():
    "A method that is not typed"
    s = ast_lambda("e.Jetsss('default')")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.Jetsss('default')"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == Any


def test_plain_object_method():
    "A method that is not typed"
    s = ast_lambda("j.pt()")
    objs = ObjectStream[Jet](ast.Name(id="j", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "j", Jet, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("j.pt()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("j"))
    assert expr_type == float


def test_function_with_processor():
    "Define a function we can use"

    def MySqrtProcessor(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
        new_s = s.MetaData({"j": "func_stuff"})
        return new_s, a

    @func_adl_callable(MySqrtProcessor)
    def MySqrt(x: float) -> float:
        ...

    s = ast_lambda("MySqrt(2)")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()), item_type=Event)

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("MySqrt(2)"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'func_stuff'})"))
    assert new_objs.item_type == Event
    assert expr_type == float


def test_function_with_simple():
    "Define a function we can use"

    @func_adl_callable()
    def MySqrt(x: float) -> float:
        ...

    s = ast_lambda("MySqrt(2)")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("MySqrt(2)"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == float


def test_function_with_missing_arg():
    "Define a function we can use"

    @func_adl_callable()
    def MySqrt(my_x: float) -> float:
        ...

    s = ast_lambda("MySqrt()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    with pytest.raises(ValueError) as e:
        remap_by_types(objs, "e", Event, s)

    assert "my_x" in str(e)


def test_function_with_default():
    "Define a function we can use"

    @func_adl_callable()
    def MySqrt(x: float = 20) -> float:
        ...

    s = ast_lambda("MySqrt()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("MySqrt(20)"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == float


def test_function_with_keyword():
    "Define a function we can use"

    @func_adl_callable()
    def MySqrt(x: float = 20) -> float:
        ...

    s = ast_lambda("MySqrt(x=15)")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("MySqrt(15)"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == float


def test_remap_lambda_helper():
    "Test simple usage of helper function"
    s = cast(ast.Lambda, ast_lambda("lambda e: e.Jets('default')"))
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()), item_type=Event)

    new_objs, new_s, rtn_type = remap_from_lambda(objs, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("lambda e: e.Jets('default')"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'stuff'})"))
    assert new_objs.item_type == Event
    assert rtn_type == Iterable[Jet]


def test_remap_lambda_subclass():
    "When objectstream is another class"

    class MyStream(ObjectStream[T]):
        def __init__(self, c, item_type: Type):
            super().__init__(c, item_type)

    s = cast(ast.Lambda, ast_lambda("lambda e: e.Jets('default')"))
    objs = MyStream[Event](ast.Name(id="e", ctx=ast.Load()), Event)

    new_objs, new_s, rtn_type = remap_from_lambda(objs, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("lambda e: e.Jets('default')"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'stuff'})"))
    assert rtn_type == Iterable[Jet]



import ast
import copy
import logging
from typing import Any, Callable, Iterable, Optional, Tuple, Type, TypeVar, cast

import pytest

from func_adl import ObjectStream
from func_adl.type_based_replacement import (
    func_adl_callable,
    func_adl_callback,
    func_adl_parameterized_call,
    register_func_adl_os_collection,
    remap_by_types,
    remap_from_lambda,
)


class Track:
    def pt(self) -> float:
        ...

    def eta(self) -> float:
        ...


T = TypeVar("T")


def add_track_extra_info(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
    s_update = s.MetaData({"t": "track stuff"})
    return s_update, a


@func_adl_callback(add_track_extra_info)
class TrackStuff:
    def pt(self) -> float:
        ...

    def eta(self) -> float:
        ...


class Jet:
    def pt(self) -> float:
        ...

    def eta(self) -> float:
        ...

    def tracks(self) -> Iterable[Track]:
        ...


def ast_lambda(lambda_func: str) -> ast.Lambda:
    "Return the ast starting from the Lambda node"
    return ast.parse(lambda_func).body[0].value  # type: ignore


def add_met_extra_info(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
    s_update = s.MetaData({"j": "pxyz stuff"})
    return s_update, a


@func_adl_callback(add_met_extra_info)
class met_extra:
    def pxy(self) -> float:
        ...


def add_met_info(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
    s_update = s.MetaData({"j": "pxy stuff"})
    return s_update, a


def add_met_method_info(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
    s_update = s.MetaData({"j": "custom stuff"})
    return s_update, a


@func_adl_callback(add_met_info)
class met:
    def pxy(self) -> float:
        ...

    def isGood(self) -> bool:
        ...

    def metobj(self) -> met_extra:
        ...

    @func_adl_callback(add_met_method_info)
    def custom(self) -> float:
        ...


def add_collection(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
    """Add a collection to the object stream"""
    assert isinstance(a.func, ast.Attribute)
    if a.func.attr == "Jets":
        s_update = s.MetaData({"j": "stuff"})
        return s_update, a
    elif a.func.attr == "EventNumber":
        new_call = copy.copy(a)
        new_call.args = [ast_lambda("20")]
        return s, new_call
    else:
        return s, a


class MyIterable(Iterable[T]):
    def Last(self) -> T:
        "Return the last element in the sequence"
        ...

    def Where(self, test: Callable[[T], bool]) -> Iterable[T]:
        ...


@func_adl_callback(add_collection)
class Event:
    def Jets(self, bank: str = "default") -> Iterable[Jet]:
        ...

    def JetsIterSub(self, bank: str = "default") -> MyIterable[Jet]:
        ...

    def Jets_req(self, bank_required: str) -> Iterable[Jet]:
        ...

    def MET(self) -> met:
        ...

    def MET_noreturntype(self):
        ...

    def Tracks(self) -> Iterable[Track]:
        ...

    def TrackStuffs(self) -> Iterable[TrackStuff]:
        ...

    def EventNumber(self) -> int:
        ...

    def MyLambdaCallback(self, cb: Callable) -> int:
        ...


def return_type_test(expr: str, arg_type: type, expected_type: type):
    s = ast_lambda(expr)
    objs = ObjectStream(ast.Name(id="e", ctx=ast.Load()), arg_type)

    _, _, expr_type = remap_by_types(objs, "e", arg_type, s)
    assert expr_type == expected_type


def test_int():
    return_type_test("1", int, int)


def test_int_neg():
    return_type_test("-11", int, int)


def test_bool():
    return_type_test("False", int, bool)


def test_str():
    return_type_test('"hi"', int, str)


def test_float():
    return_type_test("1.5", int, float)


def test_any():
    return_type_test("e", Any, Any)


def test_neg_float():
    return_type_test("-1.5", int, float)


def test_add_int():
    return_type_test("e+1", int, int)


def test_add_float():
    return_type_test("e+1.5", int, float)


def test_sub_int():
    return_type_test("e-1", int, int)


def test_sub_float():
    return_type_test("e-1.5", int, float)


def test_mul_int():
    return_type_test("e*1", int, int)


def test_mul_float():
    return_type_test("e*1.5", int, float)


def test_div_int():
    return_type_test("e/2", int, float)


def test_dib_float():
    return_type_test("e/1.5", int, float)


def test_bool_expression():
    "A bool expression"
    return_type_test("1 > 2", int, bool)


def test_bool_and_expression():
    "Using and"
    return_type_test("True and True", int, bool)


def test_bool_or_expression():
    "Using and"
    return_type_test("True or True", int, bool)


def test_abs_function_int_e():
    "A call to abs with an integer"
    return_type_test("abs(e)", int, float)


def test_abs_function_int_const():
    "A call to abs with an integer"
    return_type_test("abs(-23)", int, float)


def test_abs_function_float():
    "A call to abs with an float"
    return_type_test("abs(e)", float, float)


def test_ifexpr_onetype():
    "A ? expression"
    return_type_test("1 if True else 2", int, int)


def test_ifexpr_onetype_twotype_math():
    "A ? expression"
    return_type_test("1 if True else 2.2", int, float)


def test_ifexpr_onetype_twotypes():
    "A ? expression"
    with pytest.raises(ValueError) as e:
        return_type_test('1 if True else "2.2"', int, float)
    assert "str" in str(e)


def test_subscript():
    return_type_test("e[0]", Iterable[int], int)


def test_subscript_any():
    return_type_test("e[0]", Any, Any)


def test_collection():
    "A simple collection"
    s = ast_lambda("e.Jets('default')")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.Jets('default')"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'stuff'})"))
    assert expr_type == Iterable[Jet]


def test_required_arg():
    "A simple collection"
    s = ast_lambda("e.Jets_req()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    with pytest.raises(ValueError) as e:
        remap_by_types(objs, "e", Event, s)

    assert "bank_required" in str(e)


def test_collection_with_default():
    "A simple collection"
    s = ast_lambda("e.Jets()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.Jets('default')"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'stuff'})"))
    assert expr_type == Iterable[Jet]


def test_shortcut_nested_callback():
    """When there is a simple return, like Where, make sure that lambdas
    inside the method are called"""

    s = ast_lambda("e.TrackStuffs().Where(lambda t: abs(t.pt()) > 10)")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(
        ast_lambda("e.TrackStuffs().Where(lambda t: abs(t.pt()) > 10)")
    )
    assert ast.dump(new_objs.query_ast) == ast.dump(
        ast_lambda("MetaData(e, {'t': 'track stuff'})")
    )
    assert expr_type == Iterable[TrackStuff]


def test_shortcut_2nested_callback():
    """When there is a simple return, like Where, make sure that lambdas
    inside the method are called, but double inside"""

    s = ast_lambda(
        "ds.Select(lambda e: e.TrackStuffs()).Select(lambda ts: ts.Where(lambda t: t.pt() > 10))"
    )
    objs = ObjectStream[Iterable[Event]](ast.Name(id="ds", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "ds", Iterable[Event], s)

    assert ast.dump(new_s) == ast.dump(
        ast_lambda(
            "ds.Select(lambda e: e.TrackStuffs())"
            ".Select(lambda ts: ts.Where(lambda t: t.pt() > 10))"
        )
    )
    assert ast.dump(new_objs.query_ast) == ast.dump(
        ast_lambda("MetaData(ds, {'t': 'track stuff'})")
    )
    assert expr_type == Iterable[Iterable[TrackStuff]]


def test_collection_First(caplog):
    "A simple collection"
    caplog.set_level(logging.WARNING)

    s = ast_lambda("e.Jets().First()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    _, _, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == Jet

    assert len(caplog.text) == 0


def test_collection_Custom_Method_int(caplog):
    "A custom collection method not pre-given"
    caplog.set_level(logging.WARNING)

    M = TypeVar("M")

    @register_func_adl_os_collection
    class CustomCollection(ObjectStream[M]):
        def __init__(self, a: ast.AST, item_type: Optional[Type] = None):
            super().__init__(a, item_type)

        def MyFirst(self) -> int:
            ...

    s = ast_lambda("e.Jets().MyFirst()")
    objs = CustomCollection[Event](ast.Name(id="e", ctx=ast.Load()), Event)

    _, _, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == int

    assert len(caplog.text) == 0


def test_collection_Custom_Method_multiple_args(caplog):
    "A custom collection method not pre-given"
    caplog.set_level(logging.WARNING)

    M = TypeVar("M")

    @register_func_adl_os_collection
    class CustomCollection(ObjectStream[M]):
        def __init__(self, a: ast.AST, item_type=Any):
            super().__init__(a, item_type)

        def MyFirst(self, arg1: int, arg2: int) -> int:
            ...

    s = ast_lambda("e.Jets().MyFirst(1,3)")
    objs = CustomCollection[Event](ast.Name(id="e", ctx=ast.Load()))

    _, _, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == int

    assert len(caplog.text) == 0


def test_collection_Custom_Method_default(caplog):
    "A custom collection method not pre-given"
    caplog.set_level(logging.WARNING)

    M = TypeVar("M")

    @register_func_adl_os_collection
    class CustomCollection_default(ObjectStream[M]):
        def __init__(self, a: ast.AST, item_type):
            super().__init__(a, item_type)

        def Take(self, n: int = 5) -> ObjectStream[M]:
            ...

    s = ast_lambda("e.Jets().Take()")
    objs = CustomCollection_default[Event](ast.Name(id="e", ctx=ast.Load()), Event)

    _, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == ObjectStream[Jet]
    assert ast.dump(new_s) == ast.dump(ast_lambda("e.Jets('default').Take(5)"))

    assert len(caplog.text) == 0


def test_collection_Custom_Method_Jet(caplog):
    "A custom collection method not pre-given"
    caplog.set_level(logging.WARNING)

    M = TypeVar("M")

    class CustomCollection_Jet(ObjectStream[M]):
        def __init__(self, a: ast.AST, item_type):
            super().__init__(a, item_type)

        def MyFirst(self) -> M:
            ...

    register_func_adl_os_collection(CustomCollection_Jet)

    s = ast_lambda("e.Jets().MyFirst()")
    objs = CustomCollection_Jet[Event](ast.Name(id="e", ctx=ast.Load()), Event)

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == Jet

    assert len(caplog.text) == 0


def test_collection_CustomIterable(caplog):
    "A simple collection from an iterable with its own defined terminals"
    caplog.set_level(logging.WARNING)

    s = ast_lambda("e.JetsIterSub().Last()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == Jet

    assert len(caplog.text) == 0


def test_collection_CustomIterable_fallback(caplog):
    "A simple collection from an iterable with its own defined terminals"
    caplog.set_level(logging.WARNING)

    s = ast_lambda("e.JetsIterSub().Where(lambda j: j.pt() > 10)")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == Iterable[Jet]

    assert len(caplog.text) == 0


def test_collection_lambda_not_followed(caplog):
    "Warn if a lambda is not tracked"
    caplog.set_level(logging.WARNING)

    s = ast_lambda("e.MyLambdaCallback(lambda f: True)")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == int

    assert "lambda" in caplog.text.lower()
    assert "MyLambdaCallback" in caplog.text


def test_collection_Where(caplog):
    "A simple collection"
    caplog.set_level(logging.WARNING)

    s = ast_lambda("e.Jets().Where(lambda f: True)")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == Iterable[Jet]

    assert len(caplog.text) == 0


def test_collection_Select(caplog):
    "A simple collection"
    caplog.set_level(logging.WARNING)

    s = ast_lambda("e.Jets().Select(lambda j: j.pt())")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == Iterable[float]

    assert len(caplog.text) == 0


def test_collection_Select_meta(caplog):
    "A simple collection"
    caplog.set_level(logging.WARNING)

    s = ast_lambda("e.TrackStuffs().Select(lambda t: t.pt())")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == Iterable[float]
    assert ast.dump(new_objs.query_ast) == ast.dump(
        ast_lambda("MetaData(e, {'t': 'track stuff'})")
    )

    assert len(caplog.text) == 0


def test_method_on_collection():
    "Call a method that requires some special stuff on a returend object"
    s = ast_lambda("e.MET().pxy()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET().pxy()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'pxy stuff'})"))
    assert expr_type == float


def test_method_callback():
    "Call a method that requires some special stuff on a returend object"
    s = ast_lambda("e.MET().custom()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET().custom()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(
        ast_lambda("MetaData(MetaData(e, {'j': 'pxy stuff'}), {'j': 'custom stuff'})")
    )
    assert expr_type == float


def test_method_on_collection_bool():
    "Call a method that requires some special stuff on a returend object"
    s = ast_lambda("e.MET().isGood()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    _, _, expr_type = remap_by_types(objs, "e", Event, s)

    assert expr_type == bool


def test_method_on_method_on_collection():
    "Call a method that requires some special stuff on a returend object"
    s = ast_lambda("e.MET().metobj().pxy()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()), Event)

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET().metobj().pxy()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(
        ast_lambda("MetaData(MetaData(e, {'j': 'pxy stuff'}), {'j': 'pxyz stuff'})")
    )
    assert expr_type == float


def test_method_modify_ast():
    "Call a method that requires some special stuff on a returend object"
    s = ast_lambda("e.EventNumber()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.EventNumber(20)"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == int


def test_method_with_no_return_type(caplog):
    "A simple collection"
    caplog.set_level(logging.WARNING)
    s = ast_lambda("e.MET_noreturntype().pxy()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET_noreturntype().pxy()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == Any
    assert "MET_noreturntype" in caplog.text


def test_method_with_no_prototype(caplog):
    "A simple collection"
    caplog.set_level(logging.WARNING)
    s = ast_lambda("e.MET_bogus().pxy()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET_bogus().pxy()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == Any
    assert "MET_bogus" in caplog.text


def test_math_method(caplog):
    "A simple collection"
    caplog.set_level(logging.WARNING)
    s = ast_lambda("abs(e.MET.pxy())")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert len(caplog.text) == 0


def test_method_with_no_inital_type(caplog):
    "A simple collection"
    caplog.set_level(logging.WARNING)
    s = ast_lambda("e.MET_bogus().pxy()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Any, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.MET_bogus().pxy()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == Any
    assert len(caplog.text) == 0


def test_bogus_method():
    "A method that is not typed"
    s = ast_lambda("e.Jetsss('default')")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.Jetsss('default')"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == Any


def test_plain_object_method():
    "A method that is not typed"
    s = ast_lambda("j.pt()")
    objs = ObjectStream[Jet](ast.Name(id="j", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "j", Jet, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("j.pt()"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("j"))
    assert expr_type == float


def test_function_with_processor():
    "Define a function we can use"

    def MySqrtProcessor(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
        new_s = s.MetaData({"j": "func_stuff"})
        return new_s, a

    @func_adl_callable(MySqrtProcessor)
    def MySqrt(x: float) -> float:
        ...

    s = ast_lambda("MySqrt(2)")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()), item_type=Event)

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("MySqrt(2)"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'func_stuff'})"))
    assert new_objs.item_type == Event
    assert expr_type == float


def test_function_with_simple():
    "Define a function we can use"

    @func_adl_callable()
    def MySqrt(x: float) -> float:
        ...

    s = ast_lambda("MySqrt(2)")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("MySqrt(2)"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == float


def test_function_with_missing_arg():
    "Define a function we can use"

    @func_adl_callable()
    def MySqrt(my_x: float) -> float:
        ...

    s = ast_lambda("MySqrt()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    with pytest.raises(ValueError) as e:
        remap_by_types(objs, "e", Event, s)

    assert "my_x" in str(e)


def test_function_with_default():
    "Define a function we can use"

    @func_adl_callable()
    def MySqrt(x: float = 20) -> float:
        ...

    s = ast_lambda("MySqrt()")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("MySqrt(20)"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == float


def test_function_with_default_inside():
    "A function with a default arg that is inside a select"

    @func_adl_callable()
    def MySqrt(x: float = 20) -> float:
        ...

    s = ast_lambda("e.Jets().Select(lambda j: MySqrt())")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(
        ast_lambda("e.Jets('default').Select(lambda j: MySqrt(20))")
    )
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'stuff'})"))
    assert expr_type == Iterable[float]


def test_function_with_keyword():
    "Define a function we can use"

    @func_adl_callable()
    def MySqrt(x: float = 20) -> float:
        ...

    s = ast_lambda("MySqrt(x=15)")
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", Event, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("MySqrt(15)"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("e"))
    assert expr_type == float


def test_remap_lambda_helper():
    "Test simple usage of helper function"
    s = cast(ast.Lambda, ast_lambda("lambda e: e.Jets('default')"))
    objs = ObjectStream[Event](ast.Name(id="e", ctx=ast.Load()), item_type=Event)

    new_objs, new_s, rtn_type = remap_from_lambda(objs, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("lambda e: e.Jets('default')"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'stuff'})"))
    assert new_objs.item_type == Event
    assert rtn_type == Iterable[Jet]


def test_remap_lambda_subclass():
    "When objectstream is another class"

    class MyStream(ObjectStream[T]):
        def __init__(self, c, item_type: Type):
            super().__init__(c, item_type)

    s = cast(ast.Lambda, ast_lambda("lambda e: e.Jets('default')"))
    objs = MyStream[Event](ast.Name(id="e", ctx=ast.Load()), Event)

    new_objs, new_s, rtn_type = remap_from_lambda(objs, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("lambda e: e.Jets('default')"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'j': 'stuff'})"))
    assert rtn_type == Iterable[Jet]


def test_index_callback_1arg():
    "Indexed callback - make sure arg is passed correctly"

    param_1_capture = None

    def my_callback(
        s: ObjectStream[T], a: ast.Call, param_1: str
    ) -> Tuple[ObjectStream[T], ast.Call, Type]:
        nonlocal param_1_capture
        param_1_capture = param_1
        return (s.MetaData({"k": "stuff"}), a, float)

    class TEvent:
        @func_adl_parameterized_call(my_callback)
        @property
        def info(self):
            ...

    s = ast_lambda("e.info['fork'](55)")
    objs = ObjectStream[TEvent](ast.Name(id="e", ctx=ast.Load()))

    new_objs, new_s, expr_type = remap_by_types(objs, "e", TEvent, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.info(55)"))
    assert ast.dump(new_objs.query_ast) == ast.dump(ast_lambda("MetaData(e, {'k': 'stuff'})"))
    assert expr_type == float
    assert param_1_capture == "fork"


def test_index_callback_2arg():
    "Indexed callback - make sure 2 args are passed correctly"

    param_1_capture = None

    def my_callback(
        s: ObjectStream[T], a: ast.Call, param_1: str
    ) -> Tuple[ObjectStream[T], ast.Call, Type]:
        nonlocal param_1_capture
        param_1_capture = param_1
        return (s.MetaData({"k": "stuff"}), a, float)

    class TEvent:
        @func_adl_parameterized_call(my_callback)
        @property
        def info(self):
            ...

    s = ast_lambda("e.info['fork', 22](55)")
    objs = ObjectStream[TEvent](ast.Name(id="e", ctx=ast.Load()))

    remap_by_types(objs, "e", TEvent, s)

    assert param_1_capture == ("fork", 22)


def test_index_callback_modify_ast():
    "Indexed callback - make ast can be correctly modified"

    def my_callback(
        s: ObjectStream[T], a: ast.Call, param_1: str
    ) -> Tuple[ObjectStream[T], ast.Call, Type]:
        new_a = copy.copy(a)
        assert isinstance(a.func, ast.Attribute)
        new_a.func = ast.Attribute(value=a.func.value, attr="dude", ctx=a.func.ctx)

        return (s, new_a, float)

    class TEvent:
        @func_adl_parameterized_call(my_callback)
        @property
        def info(self):
            ...

    s = ast_lambda("e.info['fork'](55)")
    objs = ObjectStream[TEvent](ast.Name(id="e", ctx=ast.Load()))

    _, new_s, _ = remap_by_types(objs, "e", TEvent, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.dude(55)"))


def test_index_callback_modify_ast_nested():
    "Indexed callback - make ast can be correctly modified when nested in a Select"

    def my_callback(
        s: ObjectStream[T], a: ast.Call, param_1: str
    ) -> Tuple[ObjectStream[T], ast.Call, Type]:
        new_a = copy.copy(a)
        assert isinstance(a.func, ast.Attribute)
        new_a.func = ast.Attribute(value=a.func.value, attr="dude", ctx=a.func.ctx)

        return (s, new_a, float)

    class MyJet:
        @func_adl_parameterized_call(my_callback)
        @property
        def info(self):
            ...

    class TEvent:
        def Jets(self) -> Iterable[MyJet]:
            ...

    s = ast_lambda("e.Jets().Select(lambda j: j.info['fork'](55))")
    objs = ObjectStream[TEvent](ast.Name(id="e", ctx=ast.Load()))

    _, new_s, _ = remap_by_types(objs, "e", TEvent, s)

    assert ast.dump(new_s) == ast.dump(ast_lambda("e.Jets().Select(lambda j: j.dude(55))"))


def test_index_callback_on_method():
    "Add the decorator to the wrong type of thing."

    param_1_capture = None

    with pytest.raises(ValueError) as e:

        def my_callback(
            s: ObjectStream[T], a: ast.Call, param_1: str
        ) -> Tuple[ObjectStream[T], ast.Call, Type]:
            nonlocal param_1_capture
            param_1_capture = param_1
            return (s.MetaData({"k": "stuff"}), a, float)

        class TEvent:
            @func_adl_parameterized_call(my_callback)
            def info(self):
                ...

    assert "info" in str(e)


def test_index_callback_bad_prop():
    "Indexed callback - make sure arg is passed correctly"

    param_1_capture = None

    def my_callback(
        s: ObjectStream[T], a: ast.Call, param_1: str
    ) -> Tuple[ObjectStream[T], ast.Call, Type]:
        nonlocal param_1_capture
        param_1_capture = param_1
        return (s.MetaData({"k": "stuff"}), a, float)

    class TEvent:
        @func_adl_parameterized_call(my_callback)
        @property
        def info(self):
            ...

    s = ast_lambda("e.infoo['fork'](55)")
    objs = ObjectStream[TEvent](ast.Name(id="e", ctx=ast.Load()))
    with pytest.raises(AttributeError) as e:
        remap_by_types(objs, "e", TEvent, s)

    assert "infoo" in str(e)
    assert "TEvent" in str(e)


def test_index_callback_prop_not_dec():
    "Indexed callback - make sure arg is passed correctly"

    class TEvent:
        @property
        def info(self):
            ...

    s = ast_lambda("e.info['fork'](55)")
    objs = ObjectStream[TEvent](ast.Name(id="e", ctx=ast.Load()))
    with pytest.raises(ValueError) as e:
        remap_by_types(objs, "e", TEvent, s)

    assert "info" in str(e)
    assert "TEvent" in str(e)


def test_index_callback_prop_index_bad():
    "Indexed callback - make sure arg is passed correctly"

    class TEvent:
        @property
        def info(self):
            ...

    s = ast_lambda("e.info['fork':'dork'](55)")
    objs = ObjectStream[TEvent](ast.Name(id="e", ctx=ast.Load()))
    with pytest.raises(ValueError) as e:
        remap_by_types(objs, "e", TEvent, s)

    assert "info" in str(e)
    assert "TEvent" in str(e)
    assert "index" in str(e)



from typing import Any, Generic, Iterable, TypeVar

import pytest

from func_adl.util_types import (
    _resolve_type,
    build_type_dict_from_type,
    get_class_name,
    get_inherited,
    get_method_and_class,
    is_iterable,
    resolve_type_vars,
    unwrap_iterable,
)


def test_is_iter_int():
    assert not is_iterable(int)


def test_is_iter_iter():
    assert is_iterable(Iterable[int])


def test_is_iter_inherited():
    class bogus(Iterable[int]):
        def other_stuff(self):
            return 5

    assert is_iterable(bogus)


def test_is_iter_inherited_generic():
    R = TypeVar("R")

    class bogus(Iterable[R]):
        def other_stuff(self):
            return 5

    assert is_iterable(bogus[int])


def test_Any():
    assert unwrap_iterable(Any) == Any


def test_iterable():
    assert unwrap_iterable(Iterable[int]) == int


def test_inherrited():
    class bogus(Iterable[float]):
        pass

    assert unwrap_iterable(bogus) == float


def test_inherrited_generic():
    T = TypeVar("T")

    class bogus(Iterable[T]):
        pass

    myc = bogus[int]

    assert unwrap_iterable(myc) == int


def test_non_iterable():
    assert unwrap_iterable(int) == Any


def test_non_iterable_obj():
    class bogus:
        pass

    assert unwrap_iterable(bogus) == Any


def test_get_inherited_int():
    assert get_inherited(int) == Any


def test_get_inherited_generic():
    T = TypeVar("T")

    class bogus(Iterable[T]):
        pass

    myc = bogus[int]

    assert get_inherited(myc) == Iterable[int]  # type: ignore


def test_get_inherited_two_levels():
    T = TypeVar("T")
    U = TypeVar("U")

    class bogus(Generic[T]):
        pass

    class bogus2(bogus[Iterable[U]]):
        pass

    myc = bogus2[int]
    assert get_inherited(myc) == bogus[Iterable[int]]


def test_get_inherited_generic_twice():
    T = TypeVar("T")

    class bogus(Iterable[T]):
        pass

    myc = bogus[int]
    assert get_inherited(myc) == Iterable[int]  # type: ignore

    myd = bogus[float]
    assert get_inherited(myd) == Iterable[float]  # type: ignore


def test_build_type_int():
    assert build_type_dict_from_type(int) == {}


def test_build_type_generic():
    T = TypeVar("T")

    class bogus(Iterable[T]):
        pass

    myc = bogus[int]

    assert build_type_dict_from_type(myc) == {"T": int}


def test_build_type_at_level():
    T = TypeVar("T")
    U = TypeVar("U")

    class bogus(Generic[T]):
        pass

    class bogus_inher(bogus[Iterable[U]]):
        pass

    myc = bogus_inher[int]

    assert build_type_dict_from_type(myc) == {"U": int}


def test_build_type_at_level_down_one():
    T = TypeVar("T")
    U = TypeVar("U")

    class bogus(Generic[T]):
        pass

    class bogus_inher(bogus[Iterable[U]]):
        pass

    myc = bogus_inher[int]

    assert build_type_dict_from_type(myc, at_class=bogus) == {"T": Iterable[int]}


def test_build_type_at_level_down_one_reversed():
    T = TypeVar("T")
    U = TypeVar("U")

    class bogus(Generic[T]):
        pass

    class bogus_inher(bogus[Iterable[U]]):
        pass

    myc = bogus[int]

    with pytest.raises(TypeError) as e:
        build_type_dict_from_type(myc, at_class=bogus_inher)

    assert "bogus[" in str(e.value)


def test_resolve_type_int():
    assert _resolve_type(int, {}) == int


def test_resolve_type_generic():
    T = TypeVar("T")

    assert _resolve_type(T, {"T": int}) == int


def test_resolve_type_generic_filled():
    assert _resolve_type(Iterable[int], {"T": int}) == Iterable[int]


def test_resolve_type_generic_not_found():
    T = TypeVar("T")

    assert _resolve_type(T, {}) is None


def test_resolve_type_nested():
    T = TypeVar("T")

    assert _resolve_type(Iterable[T], {"T": int}) == Iterable[int]


def test_resolve_type_nested_unknown():
    K = TypeVar("K")

    assert _resolve_type(Iterable[K], {"T": int}) is None


def test_resolve_type_vars():
    T = TypeVar("T")

    class bogus(Iterable[T]):
        pass

    myc = bogus[int]

    assert resolve_type_vars(T, myc) == int


def test_resolve_type_vars_with_no_match():
    T = TypeVar("T")

    class bogus(Iterable[T]):
        pass

    class bogus_inher(bogus[int]):
        pass

    myc = bogus[int]

    assert resolve_type_vars(int, myc, at_class=bogus_inher) == int


def test_resolve_type_vars_not_there_no_match():
    T = TypeVar("T")

    class bogus(Iterable[T]):
        pass

    class bogus_inher(bogus[int]):
        pass

    myc = bogus[int]

    assert resolve_type_vars(T, myc, at_class=bogus_inher) is None


def test_get_name_simple():
    assert get_class_name(int) == "int"


def test_get_name_obj():
    class bogus:
        pass

    assert get_class_name(bogus) == "bogus"


def test_get_name_template():
    assert get_class_name(Iterable[int]) == "Iterable[int]"


def test_get_method_and_class_not_there():
    class bogus:
        pass

    assert get_method_and_class(bogus, "fork") is None


def test_get_method_and_class_easy():
    class bogus:
        def fork(self):
            pass

    assert get_method_and_class(bogus, "fork") == (bogus, bogus.fork)


def test_get_method_and_class_inherrited():
    class bogus_1:
        def fork(self):
            pass

    class bogus_2(bogus_1):
        pass

    assert get_method_and_class(bogus_2, "fork") == (bogus_1, bogus_1.fork)


def test_get_method_and_class_inherrited_override():
    class bogus_1:
        def fork(self):
            pass

    class bogus_2(bogus_1):
        def __init__(self):
            self.i = 0

        def fork(self):
            self.i = self.i + 1

    assert get_method_and_class(bogus_2, "fork") == (bogus_2, bogus_2.fork)


def test_get_method_and_class_inherrited_template():
    T = TypeVar("T")

    class bogus_1(Generic[T]):
        def fork(self) -> T:
            ...

    class bogus_2(bogus_1[int]):
        pass

    assert get_method_and_class(bogus_2, "fork") == (bogus_1, bogus_1.fork)


def test_get_method_and_class_iterable():
    class bogus:
        def fork(self):
            ...

    assert get_method_and_class(Iterable[bogus], "fork") is None


def test_get_method_and_class_Any():
    assert get_method_and_class(Any, "fork") is None



import ast
from func_adl.object_stream import ObjectStream
from typing import Any, cast

import pytest
from func_adl import EventDataset, find_EventDataset


def test_cannot_create():
    with pytest.raises(Exception):
        # This should cause an abstract method error
        EventDataset()  # type: ignore


class my_event(EventDataset):
    async def execute_result_async(self, a: ast.AST) -> Any:
        return 10


class my_event_extra_args(EventDataset):
    def __init__(self):
        super().__init__()
        cast(ast.Call, self.query_ast).args.append(ast.Str(s="hi"))

    async def execute_result_async(self, a: ast.AST) -> Any:
        return 10


def test_can_create():
    my_event()


def test_string_rep():
    e = my_event()
    assert str(e) == "my_event"
    assert repr(e) == "'my_event'"


def test_eds_recovery():
    "Make sure we can get back the event dataset"
    r = my_event()
    found_node = find_EventDataset(r.query_ast)
    assert hasattr(found_node, "_eds_object")
    assert found_node._eds_object == r  # type: ignore


def test_eds_recovery_with_select():
    r = my_event()
    q = r.Select(lambda a: a + 1)
    found_node = find_EventDataset(q.query_ast)
    assert found_node._eds_object == r  # type: ignore


def test_eds_recovery_two_ds():
    r1 = my_event()
    r2 = my_event()

    q1 = r1.Select(lambda a: a + 1)
    q2 = r2.Select(lambda b: b + 1)

    q = ObjectStream(ast.BinOp(q1.query_ast, ast.Add, q2.query_ast))
    with pytest.raises(Exception) as e:
        find_EventDataset(q.query_ast)

    assert "more than one" in str(e)


def test_eds_recovery_no_root():
    q = ObjectStream(ast.BinOp(ast.Num(1), ast.Add, ast.Num(2)))
    with pytest.raises(Exception) as e:
        find_EventDataset(q.query_ast)

    assert "no root" in str(e)


def test_eds_recovery_with_odd_call():
    r = my_event()
    q = r.Select(lambda a: (lambda b: b + 1)(a))
    found_node = find_EventDataset(q.query_ast)
    assert found_node._eds_object == r  # type: ignore



from __future__ import annotations

import ast
import inspect
import sys
import tokenize
from collections import defaultdict
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union, cast

# Some functions to enable backwards compatibility.
# Capability may be degraded in older versions - particularly 3.6.
if sys.version_info >= (3, 8):  # pragma: no cover

    def as_literal(p: Union[str, int, float, bool, None]) -> ast.Constant:
        return ast.Constant(value=p, kind=None)

else:  # pragma: no cover

    def as_literal(p: Union[str, int, float, bool, None]):
        if isinstance(p, str):
            return ast.Str(p)
        elif isinstance(p, (int, float)):
            return ast.Num(p)
        elif isinstance(p, bool):
            return ast.NameConstant(p)
        elif p is None:
            return ast.NameConstant(None)
        else:
            raise ValueError(f"Unknown type {type(p)} - do not know how to make a literal!")


def as_ast(p_var: Any) -> ast.AST:
    """Convert any python constant into an ast

    Args:
        p_var   Some python variable that can be rendered with str(p_var)
                in a way the ast parse module will be able to ingest it.

    Returns:
        A python AST representing the object. For example, if a list is passed in, then
        the result will be an AST node of type ast.List.

    """
    # If we are dealing with a string, we have to special case this.
    if isinstance(p_var, str):
        p_var = f"'{p_var}'"
    a = ast.parse(str(p_var))

    # Life out the thing inside the expression.
    # assert isinstance(a.body[0], ast.Expr)
    b = a.body[0]
    assert isinstance(b, ast.Expr)
    return b.value


def function_call(function_name: str, args: List[ast.AST]) -> ast.Call:
    """
    Generate a function call to `function_name` with a list of `args`.

    Args:
        function_name   String that is the function name
        args            List of ast's, each one is an argument.
    """
    return ast.Call(ast.Name(function_name, ast.Load()), args, [])


def lambda_unwrap(lam: ast.AST) -> ast.Lambda:
    """Given an AST of a lambda node, return the lambda node. If it is buried in a module, then
    unwrap it first Python, when it parses an module, returns the lambda wrapped in a `Module` AST
    node. This gets rid of it, but is also flexible.

    Args:
        lam:     Lambda AST. It may be wrapped in a module as well.

    Returns:
        `Lambda` AST node, regardless of how the `Lambda` was wrapped in `lam`.

    Exceptions:
        If the AST node isn't a lambda or a module wrapping a lambda.
    """
    lb = cast(ast.Expr, lam.body[0]).value if isinstance(lam, ast.Module) else lam
    if not isinstance(lb, ast.Lambda):
        raise Exception(
            f"Attempt to get lambda expression body from {type(lam)}, " "which is not a lambda."
        )

    return lb


def lambda_args(lam: Union[ast.Module, ast.Lambda]) -> ast.arguments:
    "Return the arguments of a lambda, no matter what form the lambda is in."
    return lambda_unwrap(lam).args


def lambda_body(lam: Union[ast.Lambda, ast.Module]) -> ast.AST:
    """
    Given an AST lambda node, get the expression it uses and return it. This just makes life
    easier, no real logic is occurring here.
    """
    return lambda_unwrap(lam).body


def lambda_call(args: Union[str, List[str]], lam: Union[ast.Lambda, ast.Module]) -> ast.Call:
    """
    Create a `Call` AST that calls a lambda with the named args.

    Args:
        args:       a single string or a list of strings, each string is an argument name to be
                    passed in.
        lam:        The lambda we want to call.

    Returns:
        A `Call` AST that calls the lambda with the given arguments.
    """
    if isinstance(args, str):
        args = [args]
    named_args = [ast.Name(x, ast.Load()) for x in args]
    return ast.Call(lambda_unwrap(lam), named_args, [])


if sys.version_info >= (3, 9):

    def lambda_build(args: Union[str, List[str]], l_expr: ast.AST) -> ast.Lambda:
        """
        Given a named argument(s), and an expression, build a `Lambda` AST node.

        Args:
            args:       the string names of the arguments to the lambda. May be a list or a
                        single name
            l_expr:     An AST node that is the body of the lambda.

        Returns:
            The `Lambda` AST node.
        """
        if type(args) is str:
            args = [args]

        ast_args = ast.arguments(
            posonlyargs=[],
            args=[ast.arg(arg=x) for x in args],
            kwonlyargs=[],
            kw_defaults=[],
            defaults=[],
        )
        call_lambda = ast.Lambda(args=ast_args, body=l_expr)

        return call_lambda

elif sys.version_info >= (3, 8):  # pragma: no cover

    def lambda_build(args: Union[str, List[str]], l_expr: ast.AST) -> ast.Lambda:
        """
        Given a named argument(s), and an expression, build a `Lambda` AST node.

        Args:
            args:       the string names of the arguments to the lambda. May be a list or a
                        single name
            l_expr:     An AST node that is the body of the lambda.

        Returns:
            The `Lambda` AST node.
        """
        if type(args) is str:
            args = [args]

        ast_args = ast.arguments(
            posonlyargs=[],
            vararg=None,
            args=[ast.arg(arg=x, annotation=None, type_comment=None) for x in args],
            kwonlyargs=[],
            kw_defaults=[],
            kwarg=None,
            defaults=[],
        )
        call_lambda = ast.Lambda(args=ast_args, body=l_expr)

        return call_lambda

else:  # pragma: no cover

    def lambda_build(args: Union[str, List[str]], l_expr: ast.AST) -> ast.Lambda:
        """
        Given a named argument(s), and an expression, build a `Lambda` AST node.

        Args:
            args:       the string names of the arguments to the lambda. May be a list or a
                        single name
            l_expr:     An AST node that is the body of the lambda.

        Returns:
            The `Lambda` AST node.
        """
        if type(args) is str:
            args = [args]

        ast_args = ast.arguments(
            vararg=None,
            args=[ast.arg(arg=x, annotation=None) for x in args],
            kwonlyargs=[],
            kw_defaults=[],
            kwarg=None,
            defaults=[],
        )
        call_lambda = ast.Lambda(args=ast_args, body=l_expr)

        return call_lambda


def lambda_body_replace(lam: ast.Lambda, new_expr: ast.AST) -> ast.Lambda:
    """
    Return a new lambda function that has new_expr as the body rather than the old one.
    Otherwise, everything is the same.

    Args:
        lam:        A ast.Lambda or ast.Module that points to a lambda.
        new_expr:   Expression that should replace this one.

    Returns:
        new_lam: New lambda that looks just like the old one, other than the expression is new. If
        the old one was an ast.Module, so will this one be.
    """
    if type(lam) is not ast.Lambda:
        raise Exception(
            f"Attempt to get lambda expression body from {type(lam)}, " "which is not a lambda."
        )

    new_lam = ast.Lambda(lam.args, new_expr)
    return new_lam


def lambda_assure(east: ast.AST, nargs: Optional[int] = None):
    r"""
    Make sure the Python expression ast is a lambda call, and that it has the right number of args.

    Args:
        east:        python expression ast (module ast)
        nargs:      number of args it is required to have. If None, no check is done.
    """
    if not lambda_test(east, nargs):
        raise Exception(
            "Expression AST is not a lambda function with the right number of arguments"
        )

    return east


def lambda_is_identity(lam: ast.AST) -> bool:
    "Return true if this is a lambda with 1 argument that returns the argument"
    if not lambda_test(lam, 1):
        return False

    b = lambda_unwrap(lam)
    if not isinstance(b.body, ast.Name):
        return False

    a1 = b.args.args[0].arg
    return a1 == b.body.id


def lambda_is_true(lam: ast.AST) -> bool:
    "Return true if this lambda always returns true"
    if not lambda_test(lam):
        return False
    rl = lambda_unwrap(lam)
    if not isinstance(rl.body, ast.NameConstant):
        return False

    return rl.body.value is True


def lambda_test(lam: ast.AST, nargs: Optional[int] = None) -> bool:
    r"""Test arguments"""
    if not isinstance(lam, ast.Lambda):
        if not isinstance(lam, ast.Module):
            return False
        if len(lam.body) != 1:
            return False
        if not isinstance(lam.body[0], ast.Expr):
            return False
        if not isinstance(cast(ast.Expr, lam.body[0]).value, ast.Lambda):
            return False
    rl = lambda_unwrap(lam) if type(lam) is ast.Module else lam
    if type(rl) is not ast.Lambda:
        return False
    if nargs is None:
        return True
    return len(lambda_unwrap(lam).args.args) == nargs


def rewrite_func_as_lambda(f: ast.FunctionDef) -> ast.Lambda:
    """Rewrite a function definition as a lambda. The function can contain only
    a single return statement. ValueError is throw otherwise.

    Args:
        f (ast.FunctionDef): The ast pointing to the function definition.

    Raises:
        ValueError: An error occurred during the conversion:

    Returns:
        ast.Lambda: A lambda that is the equivalent.

    Notes:

        - It is assumed that the ast passed in won't be altered in place - no deep copy is
          done of the statement or args - they are just re-used.
    """
    if len(f.body) != 1:
        raise ValueError(
            f'Can handle simple functions of only one line - "{f.name}"' f" has {len(f.body)}."
        )
    if not isinstance(f.body[0], ast.Return):
        raise ValueError(
            f'Simple function must use return statement - "{f.name}" does ' "not seem to."
        )

    # the arguments
    args = f.args
    ret = cast(ast.Return, f.body[0])
    return ast.Lambda(args, ret.value)


class _rewrite_captured_vars(ast.NodeTransformer):
    def __init__(self, cv: inspect.ClosureVars):
        self._lookup_dict: Dict[str, Any] = dict(cv.nonlocals)
        self._lookup_dict.update(cv.globals)
        self._ignore_stack = []

    def visit_Name(self, node: ast.Name) -> Any:
        if self.is_arg(node.id):
            return node

        if node.id in self._lookup_dict:
            v = self._lookup_dict[node.id]
            if not callable(v):
                return as_literal(self._lookup_dict[node.id])
        return node

    def visit_Lambda(self, node: ast.Lambda) -> Any:
        self._ignore_stack.append([a.arg for a in node.args.args])
        v = super().generic_visit(node)
        self._ignore_stack.pop()
        return v

    def is_arg(self, a_name: str) -> bool:
        "If the arg is on the stack, then return true"
        return any([a == a_name for frames in self._ignore_stack for a in frames])


def global_getclosurevars(f: Callable) -> inspect.ClosureVars:
    """Grab the closure over all passed function. Add all known global
    variables in as well.

    Args:
        f (Callable): The function pointer

    Returns:
        inspect.ClosureVars: Standard thing returned from `inspect.getclosurevars`,
            with the global variables added into it.
    """
    cv = inspect.getclosurevars(f)

    # Now, add all the globals that we might know about in case they are also
    # referenced inside this a nested lambda.
    cv.globals.update(f.__globals__)  # type: ignore

    return cv


def _realign_indent(s: str) -> str:
    """Move the first line to be at zero indent, and then apply same for everything
    below.

    Args:
        s (str): The string with indents

    Returns:
        str: Unindent string
    """
    lines = s.split("\n")
    spaces = len(lines[0]) - len(lines[0].lstrip())
    stripped_lines = [ln[spaces:] for ln in lines]
    while len(stripped_lines) > 0 and stripped_lines[-1].strip() == "":
        stripped_lines.pop()
    return "\n".join(stripped_lines)


def _get_sourcelines(f: Callable) -> Tuple[List[str], int]:
    """Get the source lines for a function, including a lambda, and make sure
    to return all the lines that might be important.

    TODO: WHen we hit Python 3.11, the better error return info might mean that tokens
    are better known and we can go back to the get source - and it will be a lot better.

    Args:
        f (Callable): The function

    Returns:
        List[str]: The source lines
    """
    lines, start_line = inspect.findsource(f)
    return lines, start_line


class _line_string_reader:
    def __init__(self, lines: List[str], initial_line: int):
        """Create a readline like interface for a list of strings.
        Uses the usual readline semantics, so that the cursor is
        always at the end of the string.

        Args:
            lines (List[str]): All the lines of the file
            initial_line (int): The next line to be read
        """
        self._lines = lines
        self._current_line = initial_line

    def readline(self) -> str:
        """Read the next line from the array. If we are off the
        end, return None

        Returns:
            Optional[str]: The next line in the file.
        """
        if self._current_line >= len(self._lines):
            return ""
        else:
            self._current_line += 1
            return self._lines[self._current_line - 1]


class _token_runner:
    def __init__(self, source: List[str], initial_line: int):
        """Track the tokenizer stream - burying a few details so higher level
        can work better.

        Args:
            source (List[str]): The source code
            initial_line (int): The initial line
        """
        self._source = source
        self._initial_line = initial_line

        self.init_tokenizer()

    def init_tokenizer(self):
        """Restart the tokenizer from the source and initial line"""
        self._tokenizer = tokenize.generate_tokens(
            _line_string_reader(self._source, self._initial_line).readline
        )

    def find_identifier(
        self, identifier: List[str], can_encounter_newline=True
    ) -> Tuple[Optional[tokenize.TokenInfo], Optional[tokenize.TokenInfo]]:
        """Find the next instance of the identifier. Return the token
        before it, and the identifier token.

        Args:
            identifier (str): The identifier to find
            can_encounter_newline (bool, optional): Can we encounter a newline. Defaults to True.

        Returns:
            Tuple[Optional[tokenize.TokenInfo], Optional[tokenize.TokenInfo]]: The tokens before
            and after the identifier. None if no `identifier` is found
        """
        last_identifier = None
        for t in self._tokenizer:
            if t.type == tokenize.NAME:
                if t.string in identifier:
                    return last_identifier, t
                last_identifier = t
            if t.type == tokenize.NEWLINE and not can_encounter_newline:
                break
        return None, None

    def tokens_till(
        self, stop_condition: Dict[int, List[str]]
    ) -> Generator[tokenize.Token, None, None]:  # type: ignore
        """Yield tokens until we find a stop condition.

        * Properly tracks parentheses, etc.

        Args:
            stop_condition (Dict[int, List[str]]): The token and string when we stop.
            can_encounter_newline (bool, optional): Can we encounter a newline. Defaults to True.

        Returns:
            Generator[tokenize.Token]: The tokens
        """
        # Keep track of the number of open parens, etc.
        parens = 0
        brackets = 0
        braces = 0

        for t in self._tokenizer:
            if (
                t.type in stop_condition
                and t.string in stop_condition[t.type]
                and parens == 0
                and brackets == 0
                and braces == 0
            ):
                return

            # Track things that could fool us
            if t.type == tokenize.OP:
                if t.string == "(":
                    parens += 1
                elif t.string == ")":
                    parens -= 1
                elif t.string == "[":
                    brackets += 1
                elif t.string == "]":
                    brackets -= 1
                elif t.string == "{":
                    braces += 1
                elif t.string == "}":
                    braces -= 1

            # Ignore comments
            if t.type == tokenize.COMMENT:
                continue

            yield t


def _get_lambda_in_stream(
    t_stream, start_token: tokenize.TokenInfo
) -> Tuple[Optional[ast.Lambda], bool]:
    """Finish of looking for a lambda in a token stream. Return the compiled
    (into an ast) lambda, and whether or not we saw a newline as we were parsing.

    Args:
        t_stream (generator): Tokenizer stream
        start_token (tokenize.TokenInfo): The starting lambda token

    Returns:
        Tuple[Optional[ast.Lambda], bool]: The compiled into an ast lambda, and whether or
        not we saw a newline
    """

    # Now, move forward until we find the end of the lambda expression.
    # We are expecting this lambda as part of a argument, so we are going to look
    # for a comma or a closing paren.
    accumulated_tokens = [start_token]
    saw_new_line = False
    for t in t_stream.tokens_till({tokenize.OP: [",", ")"]}):
        accumulated_tokens.append(t)
        if t.type == tokenize.NEWLINE or t.string == "\n":
            saw_new_line = True

    function_source = "(" + tokenize.untokenize(accumulated_tokens).lstrip() + ")"
    a_module = ast.parse(function_source)
    lda = next(
        (node for node in ast.walk(a_module) if isinstance(node, ast.Lambda)),
        None,
    )
    return lda, saw_new_line


def _parse_source_for_lambda(
    ast_source: Callable, caller_name: Optional[str] = None
) -> Optional[ast.Lambda]:
    """Use the python tokenizer to scan the source around `lambda_line`
    for a lambda. Turn that into an ast, and return it.

    Args:
        source (List[str]): Source code file
        lambda_line (int): Line in the source code where the `lambda` is seen.

    Returns:
        Optional[ast.Lambda]: The ast if a lambda is found.
    """
    # Find the start of the lambda or the separate 1-line function. We need to find the
    # enclosing function for the lambda - as funny things can be done with indents
    # and function arguments, and the tokenizer does not take kindly to surprising
    # "un-indents".
    func_name = None
    start_token = None
    source, lambda_line = _get_sourcelines(ast_source)
    t_stream = None
    while func_name is None:
        # Setup the tokenizer
        t_stream = _token_runner(source, lambda_line)

        func_name, start_token = t_stream.find_identifier(["def", "lambda"])

        if start_token is None:
            return None
        if start_token.string == "def":
            break
        if func_name is None:
            lambda_line -= 1

    assert start_token is not None
    assert t_stream is not None

    # If this is a function, then things are going to be very easy.
    if start_token.string == "def":
        function_source = _realign_indent(inspect.getsource(ast_source))
        a_module = ast.parse(function_source)
        lda = rewrite_func_as_lambda(a_module.body[0])  # type: ignore
    else:
        # Grab all the lambdas on a single line
        lambdas_on_a_line = defaultdict(list)
        saw_new_line = False
        while not saw_new_line:
            lda, saw_new_line = _get_lambda_in_stream(t_stream, start_token)
            lambdas_on_a_line[func_name.string if func_name is not None else None].append(lda)

            if saw_new_line:
                break

            func_name, start_token = t_stream.find_identifier(
                ["lambda"], can_encounter_newline=False
            )
            if start_token is None:
                break

        # Now lets make sure we pick up the right lambda. Do this by matching the caller
        # name (if we have it), and then by matching the arguments.

        lambdas_to_search = (
            lambdas_on_a_line[caller_name]
            if caller_name is not None
            else [ls for lambda_list in lambdas_on_a_line.values() for ls in lambda_list]
        )
        if len(lambdas_to_search) == 0:
            if caller_name is None:
                raise ValueError("Internal Error - Found no lambda!")
            else:
                raise ValueError(
                    f"Internal Error - Found no lambda in arguments to {caller_name}!"
                )

        def lambda_arg_list(lda: ast.Lambda) -> List[str]:
            return [a.arg for a in lda.args.args]

        caller_arg_list = inspect.getfullargspec(ast_source).args
        good_lambdas = [
            lda for lda in lambdas_to_search if lambda_arg_list(lda) == caller_arg_list
        ]
        if len(good_lambdas) == 0:
            raise ValueError(
                f"Internal Error - Found no lambda in source with the arguments {caller_arg_list}"
            )

        if len(good_lambdas) > 1:
            raise ValueError(
                "Found multiple calls to on same line"
                + ("" if caller_name is None else f" for {caller_name}")
                + " - split the calls across "
                "lines or change lambda argument names so they are different."
            )

        lda = good_lambdas[0]

    return lda


def parse_as_ast(
    ast_source: Union[str, ast.AST, Callable], caller_name: Optional[str] = None
) -> ast.Lambda:
    r"""Return an AST for a lambda function from several sources.

    We are handed one of several things:
        - An AST that is a lambda function
        - An AST that is a pointer to a Module that wraps an AST
        - Text that contains properly formatted ast code for a lambda function.

    In all cases, return a lambda function as an AST starting from the AST top node.

    Args:
        ast_source:     An AST or text string that represents the lambda.
        caller_name:    The name of the function that the lambda is an arg to. If it
            is none, then it will attempt to scan the stack frame above to figure it
            out.

    Returns:
        An ast starting from the Lambda AST node.
    """
    if callable(ast_source):
        src_ast = _parse_source_for_lambda(ast_source, caller_name)
        if not src_ast:
            # This most often happens in a notebook when the lambda is defined in a funny place
            # and can't be recovered.
            raise ValueError(f"Unable to recover source for function {ast_source}.")

        # Since this is a function in python, we can look for lambda capture.
        call_args = global_getclosurevars(ast_source)
        return _rewrite_captured_vars(call_args).visit(src_ast)

    elif isinstance(ast_source, str):
        a = ast.parse(ast_source.strip())  # type: ignore
        return lambda_unwrap(a)

    else:
        assert isinstance(ast_source, ast.AST)
        return lambda_unwrap(ast_source)


def scan_for_metadata(a: ast.AST, callback: Callable[[ast.arg], None]):
    """Scan an ast for any MetaData function calls, and pass the metadata argument
    to the call back.
    """

    class metadata_finder(ast.NodeVisitor):
        def visit_Call(self, node: ast.Call):
            self.generic_visit(node)

            if isinstance(node.func, ast.Name) and node.func.id == "MetaData":
                callback(node.args[1])  # type: ignore

    metadata_finder().visit(a)



from __future__ import annotations

import ast
import copy
import inspect
import logging
import sys
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_type_hints,
)

from .object_stream import ObjectStream
from .util_ast import as_literal, scan_for_metadata
from .util_types import (
    get_args,
    get_method_and_class,
    get_origin,
    is_iterable,
    resolve_type_vars,
    unwrap_iterable,
)

# Internal named tuple containing info for a global function
# definitions. Functions with these names can be used inline
# in func_adl expressions, and will trigger the processor and
# also be subject to normal argument resolution.
U = TypeVar("U")
_FuncAdlFunction = NamedTuple(
    "_FuncAdlFunction",
    [
        ("name", str),
        ("function", Callable),
        (
            "processor_function",
            Optional[Callable[[ObjectStream[U], ast.Call], Tuple[ObjectStream[U], ast.AST]]],
        ),
    ],
)
_global_functions: Dict[str, _FuncAdlFunction] = {}


def _load_default_global_functions():
    "Define the python standard functions that map straight through"
    # TODO: Add in other functions

    def my_abs(x: float) -> float:
        ...

    _global_functions["abs"] = _FuncAdlFunction("abs", my_abs, None)


_load_default_global_functions()


V = TypeVar("V")


def register_func_adl_function(
    function: Callable,
    processor_function: Optional[
        Callable[[ObjectStream[V], ast.Call], Tuple[ObjectStream[V], ast.AST]]
    ],
) -> None:
    """Register a new function for use inside a func_adl expression.

    * A function registered will have its arguments filled in - defaults,
    keywords converted to positional, etc.
    * If a processor function is provided, it will be called as the call
    site si being examined. Any bit can be modified.

    Args:
        function (Callable): The type definition for the function (type st)
        processor_function (Callable[[ObjectStream[T], ast.Call],
            Tuple[ObjectStream[T], ast.AST]]):
            The processor function that can modify the stream, etc.
    """
    info = _FuncAdlFunction(function.__name__, function, processor_function)
    _global_functions[info.name] = info


W = TypeVar("W")


def func_adl_callable(
    processor: Optional[
        Callable[[ObjectStream[W], ast.Call], Tuple[ObjectStream[W], ast.AST]]
    ] = None
):
    """Dectorator that will declare a function that can be used inline in
    a `func_adl` expression. The body of the function, what the backend
    translates it to, must be given by another route (e.g. via `MetaData`
    and the `processor` argument).

    ```python
    @func_adl_callable
    def my_func(arg1: float, arg2: float = 10) -> float:
        ...
    ```

    This example will declare `my_func`. If it is used like this:
    `Select(lambda x: my_func(arg1=x))`, it will be translated into
    a call sent to the backend that looks like `Select(lambda x: my_func(x, 10))`.

    If `processor` is provided, it will be called as the call site is being
    processed. One can add items to the `ObjectStream` or even modify/check the
    arguments of the call. An example:

    ```python
    def processor(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
        new_s = s.MetaData({'j': 'func_stuff'})
        return new_s, a

    @func_adl_callable(MySqrtProcessor)
    def MySqrt(x: float) -> float:
        ...
    ```

    Args:
        processor (Optional[Callable[[ObjectStream[W], ast.Call],
                  Tuple[ObjectStream[W], ast.AST]]], optional): [description]. Defaults to None.
    """
    # TODO: Do we really need to register this inside the decorator? Can we just register
    #       and return the function?
    def decorate(function: Callable):
        register_func_adl_function(function, processor)
        return function

    return decorate


def _find_keyword(
    keywords: List[ast.keyword], name: str
) -> Tuple[Optional[ast.AST], List[ast.keyword]]:
    """Find an argument in a keyword list.

    Returns:
        [type]: The argument or None if not found.
    """
    for kw in keywords:
        if kw.arg == name:
            new_kw = list(keywords)
            new_kw.remove(kw)
            return kw.value, new_kw
    return None, keywords


# Some functions to enable backwards compatibility.
# Capability may be degraded in older versions - particularly 3.6.
if sys.version_info >= (3, 8):  # pragma: no cover

    def get_type_args(tp):
        import typing

        return typing.get_args(tp)

else:  # pragma: no cover

    def get_type_args(tp):
        """Get type arguments with all substitutions performed.
        For unions, basic simplifications used by Union constructor are performed.
        Examples::
            get_args(Dict[str, int]) == (str, int)
            get_args(int) == ()
            get_args(Union[int, Union[T, int], str][int]) == (int, str)
            get_args(Union[int, Tuple[T, int]][str]) == (int, Tuple[str, int])
            get_args(Callable[[], T][int]) == ([], int)
        """
        return getattr(tp, "__args__", ())


@dataclass
class CollectionClassInfo:
    "Info for a collection class"
    obj_type: Type


_g_collection_classes: Dict[Type, CollectionClassInfo] = {}


StreamItem = TypeVar("StreamItem")


@dataclass
class PropertyCallbackInfo(Generic[StreamItem]):
    "Info for a property parameterized callback"

    # Callback
    callback: Callable[
        [ObjectStream[StreamItem], ast.Call, Any], Tuple[ObjectStream[StreamItem], ast.Call, Type]
    ]


_g_parameterized_callbacks: Dict[property, PropertyCallbackInfo] = {}


C_TYPE = TypeVar("C_TYPE")


def register_func_adl_os_collection(c: C_TYPE) -> C_TYPE:
    """Register a func_adl collections

    Args:
        c (type): Class to register

    Returns:
        [type]: [description]
    """
    _g_collection_classes[c] = CollectionClassInfo(c)
    return c


@register_func_adl_os_collection
class ObjectStreamInternalMethods(ObjectStream[StreamItem]):
    """There are a number of methods and manipulations that are not
    available at the event level, but are at the collection level. For default
    behavior, we collect those here.

    If you have a custom type for your collections, it would probably be easiest to
    in inherit from this and add your own custom methods on top of this.

    Note that these methods are *never* called. The key part of this is following
    the `item_type` through!! This is because this package does not yet know how
    to follow generics in python at runtime (think of this as poor man's type resolution).
    """

    def __init__(self, a: ast.AST, item_type: Type):
        super().__init__(a, item_type)

    @property
    def item_type(self) -> Type:
        return self._item_type

    def First(self) -> StreamItem:
        return self.item_type

    def Count(self) -> int:
        ...

    # TODO: Add all other items that belong here


def _load_g_collection_classes():
    "Use for testing to reset a global collection"
    register_func_adl_os_collection(ObjectStreamInternalMethods)


def func_adl_callback(
    callback: Callable[
        [ObjectStream[StreamItem], ast.Call], Tuple[ObjectStream[StreamItem], ast.Call]
    ]
):
    """Decorator to use on either classes or class methods that participate
    in `func_adl` queries. As these classes are traversed by the `func-adl`
    query, the given callback will be called.

    This allows the system to inject `MetaData` and other things into the query
    stream modifying the query on the fly.

    - If applied at the class level, the call back will be invoked any time any
      method is invoked on this class
    - If applied at the method level, when the method is invoked it will be applied.
    - If applied at both levels, the class level callback will be invoked first.

    Args:
        callback (Callable[[ObjectStream[StreamItem], ast.Call],
            Tuple[ObjectStream[StreamItem], ast.Call]]): Callback when object or function is used.
    """

    def decorator(cls: C_TYPE) -> C_TYPE:
        cls._func_adl_type_info = callback  # type: ignore
        return cls

    return decorator


TProp = TypeVar("TProp")


def func_adl_parameterized_call(
    cb: Callable[
        [ObjectStream[StreamItem], ast.Call, Any], Tuple[ObjectStream[StreamItem], ast.Call, Type]
    ]
) -> Callable[[TProp], TProp]:
    """Mark a property access to be treated as a parameterized callback, with `cb` being called
    when the callback is detected. The parameters from the slice operations are passed to the
    callback, but removed from the AST.

    ```python
    jet.getAttribute[cpp_float]('my_attrib')
    ```

    Gets transformed to `jet.getAttribute('my_attrib')`, and then `cb` is called with the current
    stream object (so `MetaData` can be added), the transformed ast (so it can be modified if
    desired), and `cpp_float`.

    The callback must return the updated stream, the updated (or not) ast, and the resulting type
    from the call (like `float`).

    Args:
        cb (Callable[ [ObjectStream[StreamItem], ast.Call, Any],
            Tuple[ObjectStream[StreamItem], ast.Call, Type] ]):
            The callback to be invoked when this node is found in the ast.

    Returns:
        Callable[[TProp], TProp]: Decorator function to capture a property object.
    """

    def decorator(c: TProp) -> TProp:
        if not isinstance(c, property):
            raise ValueError(f"Parameterized methods must start with proper definition ({c})")
        _g_parameterized_callbacks[c] = PropertyCallbackInfo(cb)
        return c

    return decorator


def _fill_in_default_arguments(func: Callable, call: ast.Call) -> Tuple[ast.Call, Type]:
    """Given a call and the function definition:

    * Defaults are filled in
    * A keyword argument that is positional is moved to the end
    * A follower is left to the original to help with recovering modifications to
      nested expressions.

    # TODO: Use python's signature bind to do this:
    #   https://docs.python.org/3/library/inspect.html#inspect.Signature.bind)

    Args:
        func (Callable): The function definition
        call (ast.Call): The ast call site to be modified

    Raises:
        ValueError: Missing arguments, etc.

    Returns:
        Tuple[ast.Call, Type]: The modified call site and return type.
    """
    sig = inspect.signature(func)
    i_arg = 0
    arg_array = list(call.args)
    keywords = list(call.keywords)
    for param in sig.parameters.values():
        if param.name != "self":
            if len(arg_array) <= i_arg:
                # See if they specified it as a keyword
                a, keywords = _find_keyword(keywords, param.name)
                if a is not None:
                    arg_array.append(a)  # type: ignore
                elif param.default is not param.empty:
                    a = as_literal(param.default)
                    arg_array.append(a)
                else:
                    raise ValueError(f"Argument {param.name} is required")

    # If we are making a change to the call, put in a reference back to the
    # original call.
    if len(arg_array) != len(call.args):
        old_call_ast = call
        call = copy.copy(call)
        call.args = arg_array
        call.keywords = keywords
        call._old_ast = old_call_ast  # type: ignore

    # Mark the return type - especially if it is missing
    t_info = get_type_hints(func)
    return_type = Any
    if "return" in t_info:
        return_type = t_info["return"]
    else:
        logging.getLogger(__name__).warning(
            f"Missing return annotation for {func.__name__}" " - assuming Any"
        )
        return_type = Any

    return call, return_type


def fixup_ast_from_modifications(transformed_ast: ast.AST, original_ast: ast.Call) -> ast.Call:
    "Update the ast, if needed, with modifications"

    class arg_fixer(ast.NodeVisitor):
        def __init__(self, old_copy):
            self._old_copy = old_copy
            self._new_copy = None

        @property
        def redone_ast(self) -> ast.Call:
            if self._new_copy is not None:
                return self._new_copy
            assert self._old_copy is not None
            return self._old_copy

        def _update_copy(self):
            if self._new_copy is None:
                self._new_copy = copy.copy(self._old_copy)
                self._old_copy = None

        def visit_Call(self, node: ast.Call):
            self.generic_visit(node)
            orig_ast = getattr(node, "_old_ast", None)
            if orig_ast is None:
                return node

            self._update_copy()
            n_old_args = len(orig_ast.args)
            for a in node.args[n_old_args:]:
                orig_ast.args.append(a)
            orig_ast.func = node.func

    fixer = arg_fixer(original_ast)
    fixer.visit(transformed_ast)
    return fixer.redone_ast


@dataclass
class _MethodObjectCandidate:
    """Candidate object for a particular call"""

    # The object type the method is actually getting called on
    obj_type: type

    # The object the method is defined on
    # In case of inheritance this is different than above
    method_class: type

    # The method itself
    method: Callable


@dataclass
class _MethodTypeReturnInfo:
    """For a particular method call, the return info"""

    # Update version of the ast
    node: ast.Call

    # Return type
    return_type: Type

    # If full type resolution was done (e.g. lambda following), or
    # if there were no lambda arguments to follow.
    full_type_resolution: bool

    # The object we actually did the call against
    obj_info: Optional[_MethodObjectCandidate]


T = TypeVar("T")


def remap_by_types(
    o_stream: ObjectStream[T], var_name: str, var_type: Any, a: ast.AST
) -> Tuple[ObjectStream[T], ast.AST, Type]:
    """Remap a call by a type. Given the type of `var_name` it will do its best
    to follow the objects types through the expression.

    Note:
      * Complex expressions are supported, like addition, etc.
      * Method calls are well followed
      * This isn't a complete type follower, unfortunately.

    Raises:
        ValueError: Forgotten arguments, etc.

    Returns:
        ObjectStream[T]
        ast: Updated stream and call site with
        Type: The return type of the expression given. `Any` means it couldn't
    """
    S = TypeVar("S")

    class type_transformer(ast.NodeTransformer, Generic[S]):
        def __init__(self, o_stream: ObjectStream[S]):
            self._stream = o_stream
            self._found_types: Dict[Union[str, object], type] = {var_name: var_type}

        @property
        def stream(self) -> ObjectStream[S]:
            return self._stream  # type: ignore

        def lookup_type(self, name: Union[str, object]) -> Type:
            "Return the type for a node, Any if we do not know about it"
            return self._found_types.get(name, Any)

        def process_method_call_on_stream_obj(
            self,
            obj_info: CollectionClassInfo,
            call_method_name: str,
            call_node: ast.Call,
            item_type: type,
        ) -> Optional[Tuple[ast.AST, Any]]:
            """The type following and resolution system in func_adl is pretty limited.
            When it fails, we fall back to try another method - basically, someone providing
            dummy classes that implement an `ObjectStream` like interface:
                `object_stream.query_ast`
                `object_stream.item_type`

            They are expected to build new versions of themselves much like `object_stream`.

            The reason they are here is that the API for event level is different than for
            object level. For example, `First` is defined not to make sense at the event level,
            but it very much makes sense at the object level. This provides a way to track
            the types through those calls.

            Returns:
                None if we can't figure out what to do here
                (updated call site, return annotation)
                Return annotation is replaced with the original object (e.g. Iterable)
                if that is what is started out as.
            """
            # Create the object and grab the method. This is where the API
            # of the dummy object comes into play
            s = obj_info.obj_type(ast.Name("basic"), item_type)
            call_method = getattr(s, call_method_name, None)
            assert call_method is not None

            # Call it. We can only deal with a single argument here...
            if len(call_node.args) == 0:
                r = call_method()
            elif len(call_node.args) == 1:
                r = call_method(call_node.args[0])
            else:
                return None

            # Make sure we have the right type
            if hasattr(r, "query_ast"):

                def add_md(md: ast.arg):
                    self._stream = self._stream.MetaData(ast.literal_eval(md))

                scan_for_metadata(r.query_ast, add_md)
                call_node = fixup_ast_from_modifications(r.query_ast, call_node)
                return call_node, Iterable[r.item_type]  # type: ignore

            return call_node, r

        def type_follow_in_callbacks(
            self, m_name: str, call_site_info: _MethodObjectCandidate, r_node: ast.Call
        ) -> Optional[_MethodTypeReturnInfo]:
            """If this method has any callbacks, follow them to update the stream
            and ast.

            Args:
                m_name (str): The name of hte method
                call_site_info (_MethodObjectCandidate): Method information (obj, etc.)
                r_node (ast.Call): The call site in the ast

            Returns:
                Optional[_MethodTypeReturnInfo]: If we can do type following, the results
            """
            # Next, we need to figure out what the return type is. There are multiple
            # ways, unfortunately, that we can get the return-type. We might be able to
            # do simple type resolution, but we might also have to depend on call-backs to
            # figure out what is going on (think of the Select call where the lambda return
            # type tells us the type of return of Select)

            # If this is a known collection class, we can use call-backs to follow it.
            if get_origin(call_site_info.obj_type) in _g_collection_classes:
                rtn_value = self.process_method_call_on_stream_obj(
                    _g_collection_classes[get_origin(call_site_info.obj_type)],
                    m_name,
                    r_node,
                    get_args(call_site_info.obj_type)[0],
                )
                if rtn_value is not None:
                    new_a, new_return_annotation = rtn_value
                    assert isinstance(new_a, ast.Call)
                    return _MethodTypeReturnInfo(
                        node=new_a,
                        return_type=new_return_annotation,
                        full_type_resolution=True,
                        obj_info=call_site_info,
                    )

            return None

        def process_method_callbacks(self, obj_type: type, node: ast.AST, call_method) -> ast.Call:
            """Call any callbacks that the object has registered. This might change
            the ast.

            Args:
                obj_type (type): The type of the object
                node (ast.AST): The ast node of the call site
                call_method (_type_): The method. This is the method that was called

            Returns:
                ast.AST: The updated (or not) ast.
            """
            for base_obj in [obj_type, call_method]:
                attr = getattr(base_obj, "_func_adl_type_info", None)
                if attr is not None:
                    r_stream, node = attr(self.stream, node)
                    assert isinstance(node, ast.AST)
                    self._stream = r_stream

            assert isinstance(node, ast.Call)
            return node

        def process_method_call(self, node: ast.Call, obj_type: type) -> ast.AST:
            """Manage a call against a typed object.

            1. Fill in default arguments based on the signature from the object type.
            1. Insert metadata into the stream if there are call backs based on the object type.
            1. Type follow the call site to get the return type, and to make sure anything inside
               lambda's etc., get correctly type followed as well.

            Args:
                node (ast.Call): The ast node
                obj_type (type): The object type this method call is occuring against

            Returns:
                ast.AST: An updated ast that is the new method call (with default args, etc.)
            """
            # Make reference copies that we'll populate as we go
            r_node = node

            # Find all objects that this method exists on.
            base_obj_list_all = [obj_type]
            if is_iterable(obj_type):
                item_type = unwrap_iterable(obj_type)
                base_obj_list_all += [c[item_type] for c in _g_collection_classes]

            assert isinstance(r_node.func, ast.Attribute)
            m_name = r_node.func.attr

            base_obj_list: List[_MethodObjectCandidate] = []
            for bo in base_obj_list_all:
                call_method_info = get_method_and_class(bo, m_name)
                if call_method_info is not None:
                    call_method_class, call_method = call_method_info
                    base_obj_list.append(
                        _MethodObjectCandidate(
                            obj_type=bo, method_class=call_method_class, method=call_method
                        )
                    )

            # For each definition try to get typing results out of it.
            # Take the base possible one.
            return_results: List[_MethodTypeReturnInfo] = []
            for base_obj in base_obj_list:
                # Do basic static analysis without doing any call backs.
                default_args_node, return_annotation_raw = _fill_in_default_arguments(
                    base_obj.method, r_node
                )
                return_annotation = resolve_type_vars(
                    return_annotation_raw, base_obj.obj_type, at_class=base_obj.method_class
                )

                # if the static type check worked, we might be able to use this answer.
                if return_annotation is not None:
                    has_lambda_arg = any(isinstance(a, ast.Lambda) for a in default_args_node.args)
                    return_results.append(
                        _MethodTypeReturnInfo(
                            node=default_args_node,
                            return_type=return_annotation,
                            full_type_resolution=not has_lambda_arg,
                            obj_info=base_obj,
                        )
                    )

                # If we need to do full type following, then we should do that now.
                if len(return_results) == 0 or not return_results[-1].full_type_resolution:
                    r_result = self.type_follow_in_callbacks(m_name, base_obj, default_args_node)
                    if r_result is not None:
                        return_results.append(r_result)

                # If we have a good answer, we can skip all this next stuff.
                if len(return_results) > 0 and return_results[-1].full_type_resolution:
                    break

            # If we got nothing, then we really do not know what is going on.
            if len(return_results) == 0:
                if obj_type != Any:
                    self._found_types[node] = Any
                    logging.getLogger(__name__).warning(
                        f"Method {r_node.func.attr} not " f"found on object {obj_type}"
                    )
                return_results.append(
                    _MethodTypeReturnInfo(
                        node=r_node, return_type=Any, full_type_resolution=True, obj_info=None
                    )
                )

            # See if we fully processed the call. If not, and there is a lambda, then
            # we need a warning here.
            if not any(r.full_type_resolution for r in return_results):
                if any(isinstance(a, ast.Lambda) for a in r_node.args):  # type: ignore
                    logging.getLogger(__name__).warning(
                        f"Lambda argument in {m_name} was not type followed. Class"
                        f" containing {m_name} should be corrected."
                    )

            # Next, we'll do type resolution here
            best_result = return_results[-1]
            if best_result.obj_info is not None:
                best_result.node = self.process_method_callbacks(
                    best_result.obj_info.obj_type, best_result.node, best_result.obj_info.method
                )

            # We'll pick off the first one in this case.
            r_node, return_annotation = best_result.node, best_result.return_type

            # And if we have a return annotation, then we should record it!
            self._found_types[r_node] = return_annotation
            self._found_types[node] = return_annotation

            return r_node

        def process_function_call(self, node: ast.Call, func_info: _FuncAdlFunction) -> ast.AST:
            # Make reference copies that we'll populate as we go
            r_node = node

            try:
                # Deal with default arguments and the method signature
                r_node, return_annotation = _fill_in_default_arguments(func_info.function, r_node)

                # See if someone wants to process the call
                if func_info.processor_function is not None:
                    r_stream, r_node = func_info.processor_function(self.stream, r_node)
                    assert isinstance(r_node, ast.AST)
                    self._stream = r_stream

                # And if we have a return annotation, then we should record it!
                # We do it this late because we might be changing the `r_node`
                # value above!
                self._found_types[r_node] = return_annotation
                self._found_types[node] = return_annotation

                return r_node
            except Exception as e:
                raise ValueError(
                    f"Error processing function call {node} on "
                    f"function {func_info.function.__name__} ({str(e)})"
                ) from e

        def process_paramaterized_method_call(
            self,
            node: ast.Call,
            obj_type: Type,
            attr_name: str,
            slice: ast._SliceT,  # type: ignore
            func: ast.Attribute,
        ) -> ast.Call:
            "Process a obj.method[param, param, ...](args) style call"
            # Fetch property, make sure it has info attached to it
            prop = getattr(obj_type, attr_name)
            callback_info = _g_parameterized_callbacks.get(prop, None)
            if callback_info is None:
                raise ValueError(
                    f'Property "{obj_type}{attr_name}" was not decorated with '
                    "func_adl_parameterize d_call. Malformed object or usage."
                )

            # Get the parameters from the subscript
            if isinstance(slice, ast.Index):
                parameters = ast.literal_eval(slice.value)  # type: ignore
            else:
                parameters = ast.literal_eval(slice)

            # rebuild the expression, removing the slice operation and turning this into a
            # "normal" call.
            t_node = ast.Call(func, node.args, node.keywords)

            # Run the callback processing
            r_stream, r_node, return_type = callback_info.callback(self.stream, t_node, parameters)
            self._stream = r_stream

            # Mark this ast with the fact we've updated it.
            # TODO: Make sure this is done (and tested) for all aspects of the different
            #       types of function calls (functions, regular methods, etc.)
            r_node._old_ast = node  # type: ignore

            self._found_types[r_node] = return_type
            self._found_types[node] = return_type

            return r_node

        def visit_Call(self, node: ast.Call) -> ast.AST:
            t_node = self.generic_visit(node)
            assert isinstance(t_node, ast.Call)
            if isinstance(t_node.func, ast.Attribute):
                # Do we know the type of the value?
                found_type = self.lookup_type(t_node.func.value)
                if found_type is not None:
                    t_node = self.process_method_call(t_node, found_type)
            elif isinstance(t_node.func, ast.Name):
                if t_node.func.id in _global_functions:
                    t_node = self.process_function_call(t_node, _global_functions[t_node.func.id])
            elif isinstance(t_node.func, ast.Subscript):
                if isinstance(t_node.func.value, ast.Attribute):
                    found_type = self.lookup_type(t_node.func.value.value)
                    if found_type is not None:
                        t_node = self.process_paramaterized_method_call(
                            t_node,
                            found_type,
                            t_node.func.value.attr,
                            t_node.func.slice,
                            t_node.func.value,
                        )
            return t_node

        def visit_Lambda(self, node: ast.Lambda) -> Any:
            "Prevent looking into a lambda until we actually call it"
            self._found_types[node] = Callable
            return node

        def visit_UnaryOp(self, node: ast.UnaryOp) -> Any:
            t_node = self.generic_visit(node)
            self._found_types[node] = self._found_types[node.operand]
            self._found_types[t_node] = self._found_types[node.operand]
            return t_node

        def visit_BinOp(self, node: ast.BinOp) -> Any:
            t_node = super().generic_visit(node)
            assert isinstance(t_node, ast.BinOp)
            t_left = self.lookup_type(t_node.left)
            t_right = self.lookup_type(t_node.right)

            if (t_left == Any) or (t_right == Any):
                self._found_types[node] = Any
                self._found_types[t_node] = Any
            elif (t_left == float) or (t_right == float):
                self._found_types[node] = float
                self._found_types[t_node] = float
            elif isinstance(node.op, ast.Div):
                self._found_types[node] = float
                self._found_types[t_node] = float
            else:
                self._found_types[node] = int
                self._found_types[t_node] = int

            return t_node

        def visit_BoolOp(self, node: ast.BoolOp) -> Any:
            t_node = super().generic_visit(node)
            self._found_types[node] = bool
            self._found_types[t_node] = bool

            return t_node

        def visit_Compare(self, node: ast.Compare) -> Any:
            t_node = self.generic_visit(node)
            self._found_types[node] = bool
            self._found_types[t_node] = bool
            return t_node

        def visit_IfExp(self, node: ast.IfExp) -> Any:
            t_node = self.generic_visit(node)
            assert isinstance(t_node, ast.IfExp)
            t_true = self.lookup_type(t_node.body)
            t_false = self.lookup_type(t_node.orelse)

            final_type = Any
            if t_true == t_false:
                final_type = t_true
            elif t_true in [int, float, Any] and t_false in [int, float, Any]:
                final_type = float
            else:
                raise ValueError(
                    f"IfExp branches have different types: {t_true} and {t_false}"
                    " - must be compatible"
                )
            self._found_types[node] = final_type
            self._found_types[t_node] = final_type
            return t_node

        def visit_Subscript(self, node: ast.Subscript) -> Any:
            t_node = self.generic_visit(node)
            assert isinstance(t_node, ast.Subscript)
            inner_type = unwrap_iterable(self.lookup_type(t_node.value))
            self._found_types[node] = inner_type
            self._found_types[t_node] = inner_type
            return t_node

        def visit_Name(self, node: ast.Name) -> ast.Name:
            if node.id in self._found_types:
                self._found_types[node] = self._found_types[node.id]
            elif node.id in _global_functions:
                self._found_types[node] = Callable
            else:
                logging.getLogger(__name__).warning(f"Unknown type for name {node.id}")
                self._found_types[node] = Any
            return node

        def visit_Constant(self, node: ast.Constant) -> Any:
            self._found_types[node] = type(node.value)
            return node

        def visit_Num(self, node: ast.Num) -> Any:  # pragma: no cover
            "3.7 compatability"
            self._found_types[node] = type(node.n)
            return node

        def visit_Str(self, node: ast.Str) -> Any:  # pragma: no cover
            "3.7 compatability"
            self._found_types[node] = str
            return node

        def visit_NameConstant(self, node: ast.NameConstant) -> Any:  # pragma: no cover
            "3.7 compatability"
            if node.value is None:
                raise ValueError("Do not know how to work with pythons None")
            self._found_types[node] = bool
            return node

    tt = type_transformer(o_stream)
    r_a = tt.visit(a)

    return tt.stream, r_a, tt.lookup_type(a)


def remap_from_lambda(
    o_stream: ObjectStream[T], l_func: ast.Lambda
) -> Tuple[ObjectStream[T], ast.Lambda, Type]:
    """Helper function that will translate the contents of lambda
    function with inline methods and special functions.

    Returns:
        ObjectStream[T]
        ast.AST: Updated stream and lambda function
        Type: Return type of the lambda function, Any if not known.
    """
    assert len(l_func.args.args) == 1
    orig_type = o_stream.item_type
    var_name = l_func.args.args[0].arg
    stream, new_body, return_type = remap_by_types(o_stream, var_name, orig_type, l_func.body)
    return stream, ast.Lambda(l_func.args, new_body), return_type


def reset_global_functions():
    """Resets all the global functions we know about.

    Generally only used between tests.
    """
    global _global_functions, _g_collection_classes, _g_parameterized_callbacks
    _global_functions = {}
    _load_default_global_functions()
    _g_collection_classes = {}
    _load_g_collection_classes()
    _g_parameterized_callbacks = {}



from __future__ import annotations

import ast
import logging
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from make_it_sync import make_sync

from func_adl.util_types import unwrap_iterable

from .util_ast import as_ast, function_call, parse_as_ast

# Attribute that will be used to store the executor reference
executor_attr_name = "_func_adl_executor"

T = TypeVar("T")
S = TypeVar("S")


class ReturnedDataPlaceHolder:
    """Type returned for awkward, etc.,"""

    pass


def _local_simplification(a: ast.Lambda) -> ast.Lambda:
    """Simplify the AST by removing unnecessary statements and
    syntatic sugar
    """
    from func_adl.ast.syntatic_sugar import resolve_syntatic_sugar

    r = resolve_syntatic_sugar(a)
    assert isinstance(r, ast.Lambda), "resolve_syntatic_sugar must return a lambda"
    return r


class ObjectStream(Generic[T]):
    r"""
    The objects can be events, jets, electrons, or just floats, or arrays of floats.

    `ObjectStream` holds onto the AST that will produce this stream of objects. The chain
    of `ObjectStream` objects, linked together, is a DAG that stores the user's intent.

    Every stream has an _object type_. This is the type of the elements of the stream. For example,
    the top stream, the objects are of type `Event` (or `xADOEvent`). If you transform an `Event`
    into a list of jets, then the object type will be a list of `Jet` objects. Each element of the
    stream is an array. You can also lift this second array of `Jets` and turn it into a plain
    stream of `Jets` using the `SelectMany` method below. In that case, you'll no longer be able
    to tell the boundary between events.

    `TypeVar` T is the item type (so this represents, if one were to think of this as loop-a-ble,
    `Iterable[T]`).
    """

    def __init__(self, the_ast: ast.AST, item_type: Type = Any):
        r"""
        Initialize the stream with the ast that will produce this stream of objects.
        The user will almost never use this initializer.
        """
        self._q_ast = the_ast
        self._item_type = item_type

    @property
    def item_type(self) -> Type:
        "Returns the type of the item this is a stream of. None if not known."
        return self._item_type

    @property
    def query_ast(self) -> ast.AST:
        """Return the query `ast` that this `ObjectStream` represents

        Returns:
            ast.AST: The python `ast` that is represented by this query
        """
        return self._q_ast

    def SelectMany(
        self, func: Union[str, ast.Lambda, Callable[[T], Iterable[S]]]
    ) -> ObjectStream[S]:
        r"""
        Given the current stream's object type is an array or other iterable, return
        the items in this objects type, one-by-one. This has the effect of flattening a
        nested array.

        Arguments:

            func:   The function that should be applied to this stream's objects to return
                    an iterable. Each item of the iterable is now the stream of objects.

        Returns:
            A new ObjectStream of the type of the elements.

        Notes:
            - The function can be a `lambda`, the name of a one-line function, a string that
              contains a lambda definition, or a python `ast` of type `ast.Lambda`.
        """
        from func_adl.type_based_replacement import remap_from_lambda

        n_stream, n_ast, rtn_type = remap_from_lambda(
            self, _local_simplification(parse_as_ast(func))
        )
        return ObjectStream[S](
            function_call("SelectMany", [n_stream.query_ast, cast(ast.AST, n_ast)]),
            unwrap_iterable(rtn_type),
        )

    def Select(self, f: Union[str, ast.Lambda, Callable[[T], S]]) -> ObjectStream[S]:
        r"""
        Apply a transformation function to each object in the stream, yielding a new type of
        object. There is a one-to-one correspondence between the input objects and output objects.

        Arguments:

            f:      selection function (lambda)

        Returns:

            A new ObjectStream of the transformed elements.

        Notes:
            - The function can be a `lambda`, the name of a one-line function, a string that
              contains a lambda definition, or a python `ast` of type `ast.Lambda`.
        """
        from func_adl.type_based_replacement import remap_from_lambda

        n_stream, n_ast, rtn_type = remap_from_lambda(self, _local_simplification(parse_as_ast(f)))
        return ObjectStream[S](
            function_call("Select", [n_stream.query_ast, cast(ast.AST, n_ast)]), rtn_type
        )

    def Where(self, filter: Union[str, ast.Lambda, Callable[[T], bool]]) -> ObjectStream[T]:
        r"""
        Filter the object stream, allowing only items for which `filter` evaluates as true through.

        Arguments:

            filter      A filter lambda that returns True/False.

        Returns:

            A new ObjectStream that contains only elements that pass the filter function

        Notes:
            - The function can be a `lambda`, the name of a one-line function, a string that
              contains a lambda definition, or a python `ast` of type `ast.Lambda`.
        """
        from func_adl.type_based_replacement import remap_from_lambda

        n_stream, n_ast, rtn_type = remap_from_lambda(
            self, _local_simplification(parse_as_ast(filter))
        )
        if rtn_type != bool:
            raise ValueError(f"The Where filter must return a boolean (not {rtn_type})")
        return ObjectStream[T](
            function_call("Where", [n_stream.query_ast, cast(ast.AST, n_ast)]), self.item_type
        )

    def MetaData(self, metadata: Dict[str, Any]) -> ObjectStream[T]:
        """Add metadata to the current object stream. The metadata is an arbitrary set of string
        key-value pairs. The backend must be able to properly interpret the metadata.

        Returns:
            ObjectStream: A new stream, of the same type and contents, but with metadata added.
        """
        return ObjectStream[T](
            function_call("MetaData", [self._q_ast, as_ast(metadata)]), self.item_type
        )

    def QMetaData(self, metadata: Dict[str, Any]) -> ObjectStream[T]:
        """Add query metadata to the current object stream.

        - Metadata is never transmitted to any back end
        - Metadata is per-query, not per sample.

        Warnings are issued if metadata is overwriting metadata.

        Args:
            metadata (Dict[str, Any]): Metadata to be used later

        Returns:
            ObjectStream[T]: The object stream, with metadata attached
        """
        from .ast.meta_data import lookup_query_metadata

        first = True
        base_ast = self.query_ast
        for k, v in metadata.items():
            found_md = lookup_query_metadata(self, k)
            add_md = False
            if found_md is None:
                add_md = True
            elif found_md != v:
                logging.getLogger(__name__).info(
                    f'Overwriting metadata "{k}" from its old value of "{found_md}" to "{v}"'
                )
                add_md = True
            if add_md:
                if first:
                    first = False
                    base_ast = self.MetaData({}).query_ast
                    base_ast._q_metadata = {}  # type: ignore
                base_ast._q_metadata[k] = v  # type: ignore

        return ObjectStream[T](base_ast, self.item_type)

    def AsPandasDF(
        self, columns: Union[str, List[str]] = []
    ) -> ObjectStream[ReturnedDataPlaceHolder]:
        r"""
        Return a pandas stream that contains one item, an pandas `DataFrame`.
        This `DataFrame` will contain all the data fed to it. Only non-array datatypes are
        permitted: the data must look like an Excel table.

        Arguments:

            columns     Array of names of the columns. Will default to "col0", "call1", etc.
                        Exception will be thrown if the number of columns do not match.

        """
        # To get Pandas use the ResultPandasDF function call.
        if isinstance(columns, str):
            columns = [columns]

        return ObjectStream[ReturnedDataPlaceHolder](
            function_call("ResultPandasDF", [self._q_ast, as_ast(columns)])
        )

    as_pandas = AsPandasDF

    def AsROOTTTree(
        self, filename: str, treename: str, columns: Union[str, List[str]] = []
    ) -> ObjectStream[ReturnedDataPlaceHolder]:
        r"""
        Return the sequence of items as a ROOT TTree. Each item in the ObjectStream
        will get one entry in the file. The items must be of types that the infrastructure
        can work with:

            Float               A tree with a single float in each entry will be written.
            vector<float>       A tree with a list of floats in each entry will be written.
            (<tuple>)           A tree with multiple items (leaves) will be written. Each leaf
                                must have one of the above types. Nested tuples are not supported.

        Arguments:

            filename        Name of the file in which a TTree of the objects will be written.
            treename        Name of the tree to be written to the file
            columns         Array of names of the columns. This must match the number of items
                            in a tuple to be written out.

        Returns:

            A new ObjectStream with type [(filename, treename)]. This is because multiple tree's
            may be written by the back end, and need to be concatenated together to get the full
            dataset.  The order of the files back is consistent for different queries on the same
            dataset.
        """
        if isinstance(columns, str):
            columns = [columns]

        return ObjectStream[ReturnedDataPlaceHolder](
            function_call(
                "ResultTTree", [self._q_ast, as_ast(columns), as_ast(treename), as_ast(filename)]
            )
        )

    as_ROOT_tree = AsROOTTTree

    def AsParquetFiles(
        self, filename: str, columns: Union[str, List[str]] = []
    ) -> ObjectStream[ReturnedDataPlaceHolder]:
        """Returns the sequence of items as a `parquet` file. Each item in the ObjectStream
        gets a separate entry in the file. The times must be of types that the infrastructure
        can work with:

            Float               A tree with a single float in each entry will be written.
            vector<float>       A tree with a list of floats in each entry will be written.
            (<tuple>)           A tree with multiple items (leaves) will be written. Each leaf
                                must have one of the above types. Nested tuples are not supported.
            {k:v, }             A dictionary with named columns. v is either a float or a vector
                                of floats.

        Arguments:

            filename            Name of a file in which the data will be written. Depending on
                                where the data comes from this may not be used - consider it a
                                suggestion.
            columns             If the data does not arrive by dictionary, then these are the
                                column names.

        Returns:

            A new `ObjectStream` with type `[filename]`. This is because multiple files may be
            written by the backend - the data should be concatinated together to get a final
            result. The order of the files back is consistent for different queries on the same
            dataset.
        """
        if isinstance(columns, str):
            columns = [columns]

        return ObjectStream[ReturnedDataPlaceHolder](
            function_call("ResultParquet", [self._q_ast, as_ast(columns), as_ast(filename)])
        )

    as_parquet = AsParquetFiles

    def AsAwkwardArray(
        self, columns: Union[str, List[str]] = []
    ) -> ObjectStream[ReturnedDataPlaceHolder]:
        r"""
        Return a pandas stream that contains one item, an `awkward` array, or dictionary of
        `awkward` arrays. This `awkward` will contain all the data fed to it.

        Arguments:

            columns     Array of names of the columns. Will default to "col0", "call1", etc.
                        Exception will be thrown if the number of columns do not match.

        Returns:

            An `ObjectStream` with the `awkward` array data as its one and only element.
        """
        if isinstance(columns, str):
            columns = [columns]

        return ObjectStream[ReturnedDataPlaceHolder](
            function_call("ResultAwkwardArray", [self._q_ast, as_ast(columns)])
        )

    as_awkward = AsAwkwardArray

    def _get_executor(
        self, executor: Optional[Callable[[ast.AST, Optional[str]], Awaitable[Any]]] = None
    ) -> Callable[[ast.AST, Optional[str]], Awaitable[Any]]:
        r"""
        Returns an executor that can be used to run this.
        Logic seperated out as it is used from several different places.

        Arguments:
            executor            Callback to run the AST. Can be synchronous or coroutine.

        Returns:
            An executor that is either synchronous or a coroutine.
        """
        if executor is not None:
            return executor

        # Dig into the AST until we find a node with an executor reference attached. The AST is
        # traversed by looking recursively into the source of each ObjectStream, which is always
        # the first argument in the ast.Call node.
        node = self._q_ast
        while not hasattr(node, executor_attr_name):
            node = node.args[0]  # type: ignore

        # Extract the executor from this reference.
        return getattr(node, executor_attr_name)

    async def value_async(
        self,
        executor: Optional[Callable[[ast.AST, Optional[str]], Any]] = None,
        title: Optional[str] = None,
    ) -> Any:
        r"""
        Evaluate the ObjectStream computation graph. Tracks back to the source dataset to
        understand how to evaluate the AST. It is possible to pass in an executor to override that
        behavior (used mostly for testing).

        Arguments:

            executor        A function that when called with the ast will return a future for the
                            result. If None, then uses the default executor. Normally is none
                            and the default executor specified by the `EventDatasource` is called
                            instead.
            title           Optional title to hand to the transform

        Returns

            The first element of the ObjectStream after evaluation.


        Note

            This is the non-blocking version - it will return a future which can
            be `await`ed upon until the query is done.
        """
        # Fetch the executor
        exe = self._get_executor(executor)

        # Run it
        from func_adl.ast.meta_data import remove_empty_metadata

        return await exe(remove_empty_metadata(self._q_ast), title)

    value = make_sync(value_async)
